# -*- coding: utf-8 -*-
import torch
from torch import nn, cat


class Conv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(Conv, self).__init__()
        self.conv = nn.Conv3d(in_ch, out_ch, 3, padding=1)
        self.gn = nn.GroupNorm(out_ch // 4, out_ch)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.gn(x)
        x = self.relu(x)
        return x


class DCN_Conv(nn.Module):
    def __init__(self, in_ch, out_ch, kernel_size, extend_scope, morph, if_offset, device):
        super(DCN_Conv, self).__init__()
        self.offset_conv = nn.Conv3d(in_ch, 3 * 2 * kernel_size, 3, padding=1)
        self.bn = nn.BatchNorm3d(3 * 2 * kernel_size)
        self.kernel_size = kernel_size
        self.device = device

        self.if_offset = if_offset
        self.morph = morph
        self.extend_scope = extend_scope

        self.dcn_conv_x = nn.Conv3d(in_ch, out_ch, kernel_size=(1, 1, kernel_size), stride=(1, 1, kernel_size), padding=0)  #
        self.dcn_conv_y = nn.Conv3d(in_ch, out_ch, kernel_size=(1, kernel_size, 1), stride=(1, kernel_size, 1), padding=0)  #
        self.dcn_conv_z = nn.Conv3d(in_ch, out_ch, kernel_size=(kernel_size, 1, 1), stride=(kernel_size, 1, 1), padding=0)  #

        self.dcn_conv = nn.Conv3d(in_ch, out_ch, kernel_size=kernel_size, stride=kernel_size, padding=0)
        self.gn = nn.GroupNorm(out_ch // 4, out_ch)
        self.relu = nn.ReLU(inplace=True)


    def forward(self, f):
        offset = self.offset_conv(f)
        offset = self.bn(offset)
        offset = torch.tanh(offset)
        input_shape = f.shape

        dcn = DCN(input_shape, self.kernel_size, self.extend_scope, self.morph, self.device)
        deformed_feature = dcn.deform_conv(f, offset, self.if_offset)
        if self.morph == 0:
            x = self.dcn_conv_x(deformed_feature)
            x = self.gn(x)
            x = self.relu(x)
            return x
        elif self.morph == 1:
            x = self.dcn_conv_y(deformed_feature)
            x = self.gn(x)
            x = self.relu(x)
            return x
        else:
            x = self.dcn_conv_z(deformed_feature)
            x = self.gn(x)
            x = self.relu(x)
            return x

class DCN(object):
    def __init__(self, input_shape, kernel_size, extend_scope, morph, device):
        self.num_points = kernel_size
        self.depth = input_shape[2]
        self.width = input_shape[3]
        self.height = input_shape[4]
        self.morph = morph
        self.device = device
        self.extend_scope = extend_scope  # offset (-1 ~ 1) * extend_scope
        self.num_batch = input_shape[0]   # (N,C,D,W,H)
        self.num_channels = input_shape[1]

    '''
    input: offset [N,3*K,D,W,H]
    output: [N,1,K*D,W,H]   coordinate map
    output: [N,1,K,K*W,H]   coordinate map
    output: [N,1,D,W,K*H]   coordinate map
    '''
    def _coordinate_map_3D(self, offset, if_offset):
        # offset
        offset1, offset2 = torch.split(offset, 3 * self.num_points, dim=1)
        z_offset1, y_offset1, x_offset1 = torch.split(offset1, self.num_points, dim=1)
        z_offset2, y_offset2, x_offset2 = torch.split(offset2, self.num_points, dim=1)

        z_center = torch.arange(0, self.depth).repeat([self.width*self.height])
        z_center = z_center.reshape(self.width, self.height, self.depth)
        z_center = z_center.permute(2, 1, 0)
        z_center = z_center.reshape([-1, self.depth, self.width, self.height])
        z_center = z_center.repeat([self.num_points, 1, 1, 1]).float()
        z_center = z_center.unsqueeze(0)

        y_center = torch.arange(0, self.width).repeat([self.height * self.depth])
        y_center = y_center.reshape(self.height, self.depth, self.width)
        y_center = y_center.permute(1, 2, 0)
        y_center = y_center.reshape([-1, self.depth, self.width, self.height])
        y_center = y_center.repeat([self.num_points, 1, 1, 1]).float()
        y_center = y_center.unsqueeze(0)

        x_center = torch.arange(0, self.height).repeat([self.depth * self.width])
        x_center = x_center.reshape(self.depth, self.width, self.height)
        x_center = x_center.permute(0, 1, 2)
        x_center = x_center.reshape([-1, self.depth, self.width, self.height])
        x_center = x_center.repeat([self.num_points, 1, 1, 1]).float()
        x_center = x_center.unsqueeze(0)

        if self.morph == 0:
            z = torch.linspace(0, 0, 1)
            y = torch.linspace(0, 0, 1)
            x = torch.linspace(-int(self.num_points//2), int(self.num_points//2), int(self.num_points))
            z, y, x = torch.meshgrid(z, y, x)
            z_spread = z.reshape(-1, 1)
            y_spread = y.reshape(-1, 1)
            x_spread = x.reshape(-1, 1)

            z_grid = z_spread.repeat([1, self.depth * self.width * self.height])
            z_grid = z_grid.reshape([self.num_points, self.depth, self.width, self.height])
            z_grid = z_grid.unsqueeze(0)  # [N*K,D,W,H]

            y_grid = y_spread.repeat([1, self.depth * self.width * self.height])
            y_grid = y_grid.reshape([self.num_points, self.depth, self.width, self.height])
            y_grid = y_grid.unsqueeze(0)  # [N*K*K*K,D,W,H]

            x_grid = x_spread.repeat([1, self.depth * self.width * self.height])
            x_grid = x_grid.reshape([self.num_points, self.depth, self.width, self.height])
            x_grid = x_grid.unsqueeze(0)  # [N*K*K*K,D,W,H]

            z_new = z_center + z_grid
            y_new = y_center + y_grid
            x_new = x_center + x_grid                           # [N*K*K*K,D,W,H]

            z_new = z_new.repeat(self.num_batch, 1, 1, 1, 1)
            y_new = y_new.repeat(self.num_batch, 1, 1, 1, 1)
            x_new = x_new.repeat(self.num_batch, 1, 1, 1, 1)

            z_new = z_new.to(self.device)
            y_new = y_new.to(self.device)
            x_new = x_new.to(self.device)

            z_offset1_new = z_offset1.detach().clone()
            y_offset1_new = y_offset1.detach().clone()

            if if_offset:
                z_offset1_new = z_offset1_new.permute(1, 0, 2, 3, 4)
                y_offset1_new = y_offset1_new.permute(1, 0, 2, 3, 4)
                z_offset1 = z_offset1.permute(1, 0, 2, 3, 4)
                y_offset1 = y_offset1.permute(1, 0, 2, 3, 4)
                center = int(self.num_points // 2)
                z_offset1_new[center] = 0
                y_offset1_new[center] = 0
                for index in range(1, center + 1):
                    z_offset1_new[center + index] = z_offset1_new[center + index - 1] + z_offset1[center + index]
                    z_offset1_new[center - index] = z_offset1_new[center - index + 1] + z_offset1[center - index]
                    y_offset1_new[center + index] = y_offset1_new[center + index - 1] + y_offset1[center + index]
                    y_offset1_new[center - index] = y_offset1_new[center - index + 1] + y_offset1[center - index]
                z_offset1_new = z_offset1_new.permute(1, 0, 2, 3, 4).to(self.device)
                y_offset1_new = y_offset1_new.permute(1, 0, 2, 3, 4).to(self.device)
                z_new = z_new.add(z_offset1_new.mul(self.extend_scope))
                y_new = y_new.add(y_offset1_new.mul(self.extend_scope))

                z_new = z_new.reshape([self.num_batch, 1, 1, self.num_points, self.depth, self.width, self.height])
                z_new = z_new.permute(0, 4, 1, 5, 2, 6, 3)
                z_new = z_new.reshape([self.num_batch, self.depth, self.width, self.num_points*self.height])

                y_new = y_new.reshape([self.num_batch, 1, 1, self.num_points, self.depth, self.width, self.height])
                y_new = y_new.permute(0, 4, 1, 5, 2, 6, 3)
                y_new = y_new.reshape([self.num_batch, self.depth, self.width, self.num_points*self.height])

                x_new = x_new.reshape([self.num_batch, 1, 1, self.num_points, self.depth, self.width, self.height])
                x_new = x_new.permute(0, 4, 1, 5, 2, 6, 3)
                x_new = x_new.reshape([self.num_batch, self.depth, self.width, self.num_points*self.height])
            return z_new, y_new, x_new

        elif self.morph == 1:
            z = torch.linspace(0, 0, 1)
            y = torch.linspace(-int(self.num_points // 2), int(self.num_points // 2), int(self.num_points))
            x = torch.linspace(0, 0, 1)
            z, y, x = torch.meshgrid(z, y, x)
            z_spread = z.reshape(-1, 1)
            y_spread = y.reshape(-1, 1)
            x_spread = x.reshape(-1, 1)

            z_grid = z_spread.repeat([1, self.depth * self.width * self.height])
            z_grid = z_grid.reshape([self.num_points, self.depth, self.width, self.height])
            z_grid = z_grid.unsqueeze(0)  # [N*K,D,W,H]

            y_grid = y_spread.repeat([1, self.depth * self.width * self.height])
            y_grid = y_grid.reshape([self.num_points, self.depth, self.width, self.height])
            y_grid = y_grid.unsqueeze(0)  # [N*K*K*K,D,W,H]

            x_grid = x_spread.repeat([1, self.depth * self.width * self.height])
            x_grid = x_grid.reshape([self.num_points, self.depth, self.width, self.height])
            x_grid = x_grid.unsqueeze(0)  # [N*K*K*K,D,W,H]

            z_new = z_center + z_grid
            y_new = y_center + y_grid
            x_new = x_center + x_grid  # [N*K*K*K,D,W,H]

            z_new = z_new.repeat(self.num_batch, 1, 1, 1, 1)
            y_new = y_new.repeat(self.num_batch, 1, 1, 1, 1)
            x_new = x_new.repeat(self.num_batch, 1, 1, 1, 1)

            z_new = z_new.to(self.device)
            y_new = y_new.to(self.device)
            x_new = x_new.to(self.device)
            x_offset1_new = x_offset1.detach().clone()
            z_offset2_new = z_offset2.detach().clone()

            if if_offset:
                x_offset1_new = x_offset1_new.permute(1, 0, 2, 3, 4)
                z_offset2_new = z_offset2_new.permute(1, 0, 2, 3, 4)
                x_offset1 = x_offset1.permute(1, 0, 2, 3, 4)
                z_offset2 = z_offset2.permute(1, 0, 2, 3, 4)
                center = int(self.num_points // 2)
                x_offset1_new[center] = 0
                z_offset2_new[center] = 0
                for index in range(1, center + 1):
                    x_offset1_new[center + index] = x_offset1_new[center + index - 1] + x_offset1[center + index]
                    x_offset1_new[center - index] = x_offset1_new[center - index + 1] + x_offset1[center - index]
                    z_offset2_new[center + index] = z_offset2_new[center + index - 1] + z_offset2[center + index]
                    z_offset2_new[center - index] = z_offset2_new[center - index + 1] + z_offset2[center - index]
                x_offset1_new = x_offset1_new.permute(1, 0, 2, 3, 4).to(self.device)
                z_offset2_new = z_offset2_new.permute(1, 0, 2, 3, 4).to(self.device)
                z_new = z_new.add(z_offset2_new.mul(self.extend_scope))
                x_new = x_new.add(x_offset1_new.mul(self.extend_scope))
            z_new = z_new.reshape([self.num_batch, 1, self.num_points, 1, self.depth, self.width, self.height])
            z_new = z_new.permute(0, 4, 1, 5, 2, 6, 3)
            z_new = z_new.reshape([self.num_batch, self.depth, self.num_points * self.width, self.height])
            y_new = y_new.reshape([self.num_batch, 1, self.num_points, 1, self.depth, self.width, self.height])
            y_new = y_new.permute(0, 4, 1, 5, 2, 6, 3)
            y_new = y_new.reshape([self.num_batch, self.depth, self.num_points * self.width, self.height])
            x_new = x_new.reshape([self.num_batch, 1, self.num_points, 1, self.depth, self.width, self.height])
            x_new = x_new.permute(0, 4, 1, 5, 2, 6, 3)
            x_new = x_new.reshape([self.num_batch, self.depth, self.num_points * self.width, self.height])
            return z_new, y_new, x_new

        else:
            z = torch.linspace(-int(self.num_points // 2), int(self.num_points // 2), int(self.num_points))
            y = torch.linspace(0, 0, 1)
            x = torch.linspace(0, 0, 1)
            z, y, x = torch.meshgrid(z, y, x)
            z_spread = z.reshape(-1, 1)
            y_spread = y.reshape(-1, 1)
            x_spread = x.reshape(-1, 1)

            z_grid = z_spread.repeat([1, self.depth * self.width * self.height])
            z_grid = z_grid.reshape([self.num_points, self.depth, self.width, self.height])
            z_grid = z_grid.unsqueeze(0)  # [N*K,D,W,H]

            y_grid = y_spread.repeat([1, self.depth * self.width * self.height])
            y_grid = y_grid.reshape([self.num_points, self.depth, self.width, self.height])
            y_grid = y_grid.unsqueeze(0)  # [N*K*K*K,D,W,H]

            x_grid = x_spread.repeat([1, self.depth * self.width * self.height])
            x_grid = x_grid.reshape([self.num_points, self.depth, self.width, self.height])
            x_grid = x_grid.unsqueeze(0)  # [N*K*K*K,D,W,H]

            z_new = z_center + z_grid
            y_new = y_center + y_grid
            x_new = x_center + x_grid  # [N*K*K*K,D,W,H]

            z_new = z_new.repeat(self.num_batch, 1, 1, 1, 1)
            y_new = y_new.repeat(self.num_batch, 1, 1, 1, 1)
            x_new = x_new.repeat(self.num_batch, 1, 1, 1, 1)
            
            z_new = z_new.to(self.device)
            y_new = y_new.to(self.device)
            x_new = x_new.to(self.device)
            x_offset2_new = x_offset2.detach().clone()
            y_offset2_new = y_offset2.detach().clone()

            if if_offset:
                x_offset2_new = x_offset2_new.permute(1, 0, 2, 3, 4)
                y_offset2_new = y_offset2_new.permute(1, 0, 2, 3, 4)
                x_offset2 = x_offset2.permute(1, 0, 2, 3, 4)
                y_offset2 = y_offset2.permute(1, 0, 2, 3, 4)
                center = int(self.num_points // 2)
                x_offset2_new[center] = 0
                x_offset2_new[center] = 0
                for index in range(1, center + 1):
                    x_offset2_new[center + index] = x_offset2_new[center + index - 1] + x_offset2[center + index]
                    x_offset2_new[center - index] = x_offset2_new[center - index + 1] + x_offset2[center - index]
                    y_offset2_new[center + index] = y_offset2_new[center + index - 1] + y_offset2[center + index]
                    y_offset2_new[center - index] = y_offset2_new[center - index + 1] + y_offset2[center - index]
                x_offset2_new = x_offset2_new.permute(1, 0, 2, 3, 4).to(self.device)
                y_offset2_new = y_offset2_new.permute(1, 0, 2, 3, 4).to(self.device)
                x_new = x_new.add(x_offset2_new.mul(self.extend_scope))
                y_new = y_new.add(y_offset2_new.mul(self.extend_scope))

            z_new = z_new.reshape([self.num_batch, self.num_points, 1, 1, self.depth, self.width, self.height])
            z_new = z_new.permute(0, 4, 1, 5, 2, 6, 3)
            z_new = z_new.reshape([self.num_batch, self.num_points*self.depth, self.width, self.height])

            y_new = y_new.reshape([self.num_batch, self.num_points, 1, 1, self.depth, self.width, self.height])
            y_new = y_new.permute(0, 4, 1, 5, 2, 6, 3)
            y_new = y_new.reshape([self.num_batch, self.num_points*self.depth, self.width, self.height])

            x_new = x_new.reshape([self.num_batch, self.num_points, 1, 1, self.depth, self.width, self.height])
            x_new = x_new.permute(0, 4, 1, 5, 2, 6, 3)
            x_new = x_new.reshape([self.num_batch, self.num_points*self.depth, self.width, self.height])
            return z_new, y_new, x_new

    '''
    input: input feature map [N,C,D,W,H]；coordinate map [N,K*D,K*W,K*H] 
    output: [N,1,K*D,K*W,K*H]  deformed feature map
    '''
    def _bilinear_interpolate_3D(self, input_feature, z, y, x):
        z = z.reshape([-1]).float()
        y = y.reshape([-1]).float()
        x = x.reshape([-1]).float()  # [N*KD*KW*KH]

        zero = torch.zeros([]).int()
        max_z = self.depth - 1
        max_y = self.width - 1
        max_x = self.height - 1

        # find 8 grid locations
        z0 = torch.floor(z).int()
        z1 = z0 + 1
        y0 = torch.floor(y).int()
        y1 = y0 + 1
        x0 = torch.floor(x).int()
        x1 = x0 + 1

        # clip out coordinates exceeding feature map volume以外的点
        z0 = torch.clamp(z0, zero, max_z)
        z1 = torch.clamp(z1, zero, max_z)
        y0 = torch.clamp(y0, zero, max_y)
        y1 = torch.clamp(y1, zero, max_y)
        x0 = torch.clamp(x0, zero, max_x)
        x1 = torch.clamp(x1, zero, max_x)  # [N*KD*KW*KH]

        # convert input_feature and coordinate X, Y to 3D，for gathering
        # input_feature_flat = input_feature.reshape([-1, self.num_channels])   # [N*D*W*H, C]
        input_feature_flat = input_feature.flatten()
        input_feature_flat = input_feature_flat.reshape(self.num_batch, self.num_channels, self.depth, self.width,
                                                        self.height)
        input_feature_flat = input_feature_flat.permute(0, 2, 3, 4, 1)
        input_feature_flat = input_feature_flat.reshape(-1, self.num_channels)
        dimension = self.height * self.width * self.depth

        base = torch.arange(self.num_batch) * dimension
        base = base.reshape([-1, 1]).float()  # [N,1]

        repeat = torch.ones([self.num_points * self.depth * self.width * self.height]).unsqueeze(0)
        repeat = repeat.float()  # [1,D*W*H*K*K*K]

        base = torch.matmul(base, repeat)  # [N,1] * [1,D*W*H*K*K*K]  ==> [N,D*W*H*K*K*K]
        base = base.reshape([-1])  # [D*W*H*K*K*K]

        base = base.to(self.device)

        base_z0 = base + z0 * self.height * self.width
        base_z1 = base + z1 * self.height * self.width
        base_y0 = base + y0 * self.height
        base_y1 = base + y1 * self.height

        # top rectangle of the neighbourhood volume
        index_a0 = base_y0 + base_z0 - base + x0
        index_b0 = base_y0 + base_z1 - base + x0
        index_c0 = base_y0 + base_z0 - base + x1
        index_d0 = base_y0 + base_z1 - base + x1  # [N*KD*KW*KH]

        # bottom rectangle of the neighbourhood volume
        index_a1 = base_y1 + base_z0 - base + x0
        index_b1 = base_y1 + base_z1 - base + x0
        index_c1 = base_y1 + base_z0 - base + x1
        index_d1 = base_y1 + base_z1 - base + x1  # [N*KD*KW*KH]

        # get 8 grid values  ([N*D*W*H,C], [N*D*W*H*27])
        value_a0 = input_feature_flat[index_a0.type(torch.int64)]
        value_b0 = input_feature_flat[index_b0.type(torch.int64)]
        value_c0 = input_feature_flat[index_c0.type(torch.int64)]
        value_d0 = input_feature_flat[index_d0.type(torch.int64)]
        value_a1 = input_feature_flat[index_a1.type(torch.int64)]
        value_b1 = input_feature_flat[index_b1.type(torch.int64)]
        value_c1 = input_feature_flat[index_c1.type(torch.int64)]
        value_d1 = input_feature_flat[index_d1.type(torch.int64)]  # [N*KD*KW*KH, C]

        # find 8 grid locations
        z0 = torch.floor(z).int()
        z1 = z0 + 1
        y0 = torch.floor(y).int()
        y1 = y0 + 1
        x0 = torch.floor(x).int()
        x1 = x0 + 1

        # clip out coordinates exceeding feature map volume以外的点
        z0 = torch.clamp(z0, zero, max_z + 1)
        z1 = torch.clamp(z1, zero, max_z + 1)
        y0 = torch.clamp(y0, zero, max_y + 1)
        y1 = torch.clamp(y1, zero, max_y + 1)
        x0 = torch.clamp(x0, zero, max_x + 1)
        x1 = torch.clamp(x1, zero, max_x + 1)  # [N*KD*KW*KH]

        x0_float = x0.float()
        x1_float = x1.float()
        y0_float = y0.float()
        y1_float = y1.float()
        z0_float = z0.float()
        z1_float = z1.float()

        vol_a0 = ((z1_float - z) * (y1_float - y) * (x1_float - x)).unsqueeze(-1)
        vol_b0 = ((z - z0_float) * (y1_float - y) * (x1_float - x)).unsqueeze(-1)
        vol_c0 = ((z1_float - z) * (y1_float - y) * (x - x0_float)).unsqueeze(-1)
        vol_d0 = ((z - z0_float) * (y1_float - y) * (x - x0_float)).unsqueeze(-1)
        vol_a1 = ((z1_float - z) * (y - y0_float) * (x1_float - x)).unsqueeze(-1)
        vol_b1 = ((z - z0_float) * (y - y0_float) * (x1_float - x)).unsqueeze(-1)
        vol_c1 = ((z1_float - z) * (y - y0_float) * (x - x0_float)).unsqueeze(-1)
        vol_d1 = ((z - z0_float) * (y - y0_float) * (x - x0_float)).unsqueeze(-1)  # [N*KD*KW*KH, C]

        outputs = value_a0 * vol_a0 + value_b0 * vol_b0 + value_c0 * vol_c0 + value_d0 * vol_d0 + value_a1 * vol_a1 + value_b1 * vol_b1 + value_c1 * vol_c1 + value_d1 * vol_d1

        if self.morph == 0:
            outputs = outputs.reshape([self.num_batch, self.depth, self.width, self.num_points*self.height, self.num_channels])
            outputs = outputs.permute(0, 4, 1, 2, 3)
        elif self.morph == 1:
            outputs = outputs.reshape([self.num_batch, self.depth, self.num_points*self.width, self.height, self.num_channels])
            outputs = outputs.permute(0, 4, 1, 2, 3)
        else:
            outputs = outputs.reshape([self.num_batch, self.num_points*self.depth, self.width, self.height, self.num_channels])
            outputs = outputs.permute(0, 4, 1, 2, 3)
        return outputs

    def deform_conv(self, input, offset, if_offset):
        z, y, x = self._coordinate_map_3D(offset, if_offset)
        deformed_feature = self._bilinear_interpolate_3D(input, z, y, x)
        return deformed_feature


from mmseg.registry import MODELS
from torch import nn
from mmseg.utils import ConfigType

class MMSegModel(nn.Module):
    def __init__(self, 
                 input_channels:int, 
                 num_classes:int, 
                 cfg:ConfigType = None, 
                 final_up:int =0, 
                 deep_supervision:bool = True):
        super().__init__()
        model = cfg.model
        model.data_preprocessor = None
        model.backbone.in_channels = input_channels
        model.decode_head.num_classes = num_classes
        model.auxiliary_head.num_classes = num_classes
        self.model = MODELS.build(model)
        self.final_up = nn.UpsamplingBilinear2d(scale_factor=final_up) if final_up > 0 else nn.Identity()
        self.deep_supervision = deep_supervision

    def forward(self, inputs):
        x = self.model.extract_feat(inputs)
        main_out = self.final_up(self.model.decode_head.forward(x))
        if self.deep_supervision:
            out = [main_out]
            if isinstance(self.model.auxiliary_head, nn.ModuleList):
                for idx, aux_head in enumerate(self.auxiliary_head):
                    out.append(aux_head.forward(x))
            else:
                out.append(self.model.auxiliary_head.forward(x))  
            return out
        return main_out


if __name__ == "__main__":
    import torch
    from mmengine import Config
    cfg = Config.fromfile('/home/yjiang/workspace/code/segmentation-model-zoo/smz/models/mmseg_models/configs/CCNet/ccnet_r50-d8.py')
    model = MMSegModel(input_channels=4, num_classes=3, cfg=cfg, final_up=4, deep_supervision=False)
    model.eval()
    x = torch.randn(2, 4, 256, 256)
    with torch.no_grad():
        y = model(x)
    print(y.shape)


import numpy as np
import math
import torch
from torch import nn
from torch.nn import functional as F
from typing import Union, Type, List, Tuple

from dynamic_network_architectures.building_blocks.helper import get_matching_convtransp

from torch.nn.modules.conv import _ConvNd
from torch.nn.modules.dropout import _DropoutNd
from dynamic_network_architectures.building_blocks.helper import convert_conv_op_to_dim

from nnunetv2.utilities.plans_handling.plans_handler import ConfigurationManager, PlansManager
from dynamic_network_architectures.building_blocks.helper import get_matching_instancenorm, convert_dim_to_conv_op
from dynamic_network_architectures.initialization.weight_init import init_last_bn_before_add_to_0
from nnunetv2.utilities.network_initialization import InitWeights_He
try:
    from mamba_ssm import Mamba
except ImportError:
    Mamba = None
from dynamic_network_architectures.building_blocks.helper import maybe_convert_scalar_to_list, get_matching_pool_op
from torch.amp import autocast
from dynamic_network_architectures.building_blocks.residual import BasicBlockD
from smz.registry import MODELS

class UpsampleLayer(nn.Module):
    def __init__(
            self,
            conv_op,
            input_channels,
            output_channels,
            pool_op_kernel_size,
            mode='nearest'
        ):
        super().__init__()
        self.conv = conv_op(input_channels, output_channels, kernel_size=1)
        self.pool_op_kernel_size = pool_op_kernel_size
        self.mode = mode
        
    def forward(self, x):
        x = F.interpolate(x, scale_factor=self.pool_op_kernel_size, mode=self.mode)
        x = self.conv(x)
        return x

class MambaLayer(nn.Module):
    def __init__(self, dim, d_state = 16, d_conv = 4, expand = 2, channel_token = False):
        super().__init__()
        print(f"MambaLayer: dim: {dim}")
        self.dim = dim
        self.norm = nn.LayerNorm(dim)
        self.mamba = Mamba(
                d_model=dim, # Model dimension d_model
                d_state=d_state,  # SSM state expansion factor
                d_conv=d_conv,    # Local convolution width
                expand=expand,    # Block expansion factor
        )
        self.channel_token = channel_token ## whether to use channel as tokens

    def forward_patch_token(self, x):
        B, d_model = x.shape[:2]
        assert d_model == self.dim
        n_tokens = x.shape[2:].numel()
        img_dims = x.shape[2:]
        x_flat = x.reshape(B, d_model, n_tokens).transpose(-1, -2)
        x_norm = self.norm(x_flat)
        x_mamba = self.mamba(x_norm)
        out = x_mamba.transpose(-1, -2).reshape(B, d_model, *img_dims)

        return out

    def forward_channel_token(self, x):
        B, n_tokens = x.shape[:2]
        d_model = x.shape[2:].numel()
        assert d_model == self.dim, f"d_model: {d_model}, self.dim: {self.dim}"
        img_dims = x.shape[2:]
        x_flat = x.flatten(2)
        assert x_flat.shape[2] == d_model, f"x_flat.shape[2]: {x_flat.shape[2]}, d_model: {d_model}"
        x_norm = self.norm(x_flat)
        x_mamba = self.mamba(x_norm)
        out = x_mamba.reshape(B, n_tokens, *img_dims)

        return out

    @autocast('cuda', enabled=False)
    def forward(self, x):
        if x.dtype == torch.float16:
            x = x.type(torch.float32)
        
        if self.channel_token:
            out = self.forward_channel_token(x)
        else:
            out = self.forward_patch_token(x)

        return out


class BasicResBlock(nn.Module):
    def __init__(
            self,
            conv_op,
            input_channels,
            output_channels,
            norm_op,
            norm_op_kwargs,
            kernel_size=3,
            padding=1,
            stride=1,
            use_1x1conv=False,
            nonlin=nn.LeakyReLU,
            nonlin_kwargs={'inplace': True}
        ):
        super().__init__()
        
        self.conv1 = conv_op(input_channels, output_channels, kernel_size, stride=stride, padding=padding)
        self.norm1 = norm_op(output_channels, **norm_op_kwargs)
        self.act1 = nonlin(**nonlin_kwargs)
        
        self.conv2 = conv_op(output_channels, output_channels, kernel_size, padding=padding)
        self.norm2 = norm_op(output_channels, **norm_op_kwargs)
        self.act2 = nonlin(**nonlin_kwargs)
        
        if use_1x1conv:
            self.conv3 = conv_op(input_channels, output_channels, kernel_size=1, stride=stride)
        else:
            self.conv3 = None
                  
    def forward(self, x):
        y = self.conv1(x)
        y = self.act1(self.norm1(y))  
        y = self.norm2(self.conv2(y))
        if self.conv3:
            x = self.conv3(x)
        y += x
        return self.act2(y)
    
class ResidualMambaEncoder(nn.Module):
    def __init__(self,
                 input_size: Tuple[int, ...],
                 input_channels: int,
                 n_stages: int,
                 features_per_stage: Union[int, List[int], Tuple[int, ...]],
                 conv_op: Type[_ConvNd],
                 kernel_sizes: Union[int, List[int], Tuple[int, ...]],
                 strides: Union[int, List[int], Tuple[int, ...], Tuple[Tuple[int, ...], ...]],
                 n_blocks_per_stage: Union[int, List[int], Tuple[int, ...]],
                 conv_bias: bool = False,
                 norm_op: Union[None, Type[nn.Module]] = None,
                 norm_op_kwargs: dict = None,
                 nonlin: Union[None, Type[torch.nn.Module]] = None,
                 nonlin_kwargs: dict = None,
                 return_skips: bool = False,
                 stem_channels: int = None,
                 pool_type: str = 'conv',
                 ):
        super().__init__()
        if isinstance(kernel_sizes, int):
            kernel_sizes = [kernel_sizes] * n_stages
        if isinstance(features_per_stage, int):
            features_per_stage = [features_per_stage] * n_stages
        if isinstance(n_blocks_per_stage, int):
            n_blocks_per_stage = [n_blocks_per_stage] * n_stages
        if isinstance(strides, int):
            strides = [strides] * n_stages
        assert len(
            kernel_sizes) == n_stages, "kernel_sizes must have as many entries as we have resolution stages (n_stages)"
        assert len(
            n_blocks_per_stage) == n_stages, "n_conv_per_stage must have as many entries as we have resolution stages (n_stages)"
        assert len(
            features_per_stage) == n_stages, "features_per_stage must have as many entries as we have resolution stages (n_stages)"
        assert len(strides) == n_stages, "strides must have as many entries as we have resolution stages (n_stages). " \
                                         "Important: first entry is recommended to be 1, else we run strided conv drectly on the input"

        pool_op = get_matching_pool_op(conv_op, pool_type=pool_type) if pool_type != 'conv' else None

        do_channel_token = [False] * n_stages
        feature_map_sizes = []
        feature_map_size = input_size
        for s in range(n_stages):
            feature_map_sizes.append([i // j for i, j in zip(feature_map_size, strides[s])])
            feature_map_size = feature_map_sizes[-1]
            if np.prod(feature_map_size) <= features_per_stage[s]:
                do_channel_token[s] = True
            

        print(f"feature_map_sizes: {feature_map_sizes}")
        print(f"do_channel_token: {do_channel_token}")

        self.conv_pad_sizes = []
        for krnl in kernel_sizes:
            self.conv_pad_sizes.append([i // 2 for i in krnl])

        stem_channels = features_per_stage[0]
        self.stem = nn.Sequential(
            BasicResBlock(
                conv_op = conv_op,
                input_channels = input_channels,
                output_channels = stem_channels,
                norm_op=norm_op,
                norm_op_kwargs=norm_op_kwargs,
                kernel_size=kernel_sizes[0],
                padding=self.conv_pad_sizes[0],
                stride=1,
                nonlin=nonlin,
                nonlin_kwargs=nonlin_kwargs,
                use_1x1conv=True
            ), 
            *[
                BasicBlockD(
                    conv_op = conv_op,
                    input_channels = stem_channels,
                    output_channels = stem_channels,
                    kernel_size = kernel_sizes[0],
                    stride = 1,
                    conv_bias = conv_bias,
                    norm_op = norm_op,
                    norm_op_kwargs = norm_op_kwargs,
                    nonlin = nonlin,
                    nonlin_kwargs = nonlin_kwargs,
                ) for _ in range(n_blocks_per_stage[0] - 1)
            ]
        )

        input_channels = stem_channels

        stages = []
        mamba_layers = []
        for s in range(n_stages):
            stage = nn.Sequential(
                BasicResBlock(
                    conv_op = conv_op,
                    norm_op = norm_op,
                    norm_op_kwargs = norm_op_kwargs,
                    input_channels = input_channels,
                    output_channels = features_per_stage[s],
                    kernel_size = kernel_sizes[s],
                    padding=self.conv_pad_sizes[s],
                    stride=strides[s],
                    use_1x1conv=True,
                    nonlin = nonlin,
                    nonlin_kwargs = nonlin_kwargs
                ),
                *[
                    BasicBlockD(
                        conv_op = conv_op,
                        input_channels = features_per_stage[s],
                        output_channels = features_per_stage[s],
                        kernel_size = kernel_sizes[s],
                        stride = 1,
                        conv_bias = conv_bias,
                        norm_op = norm_op,
                        norm_op_kwargs = norm_op_kwargs,
                        nonlin = nonlin,
                        nonlin_kwargs = nonlin_kwargs,
                    ) for _ in range(n_blocks_per_stage[s] - 1)
                ]
            )

            if bool(s % 2) ^ bool(n_stages % 2): ## gurantee the last stage has mamaba layer
                mamba_layers.append(
                    MambaLayer(
                        dim = np.prod(feature_map_sizes[s]) if do_channel_token[s] else features_per_stage[s],
                        channel_token = do_channel_token[s]
                    )
                )
            else:
                mamba_layers.append(nn.Identity())

            stages.append(stage)
            input_channels = features_per_stage[s]

        self.mamba_layers = nn.ModuleList(mamba_layers)
        self.stages = nn.ModuleList(stages)
        self.output_channels = features_per_stage
        self.strides = [maybe_convert_scalar_to_list(conv_op, i) for i in strides]
        self.return_skips = return_skips

        self.conv_op = conv_op
        self.norm_op = norm_op
        self.norm_op_kwargs = norm_op_kwargs
        self.nonlin = nonlin
        self.nonlin_kwargs = nonlin_kwargs
        #self.dropout_op = dropout_op
        #self.dropout_op_kwargs = dropout_op_kwargs
        self.conv_bias = conv_bias
        self.kernel_sizes = kernel_sizes

    def forward(self, x):
        if self.stem is not None:
            x = self.stem(x)
        ret = []
        for s in range(len(self.stages)):
            x = self.stages[s](x)
            x = self.mamba_layers[s](x)
            ret.append(x)
        if self.return_skips:
            return ret
        else:
            return ret[-1]

    def compute_conv_feature_map_size(self, input_size):
        if self.stem is not None:
            output = self.stem.compute_conv_feature_map_size(input_size)
        else:
            output = np.int64(0)

        for s in range(len(self.stages)):
            output += self.stages[s].compute_conv_feature_map_size(input_size)
            input_size = [i // j for i, j in zip(input_size, self.strides[s])]

        return output


class UNetResDecoder(nn.Module):
    def __init__(self,
                 encoder,
                 num_classes,
                 n_conv_per_stage: Union[int, Tuple[int, ...], List[int]],
                 deep_supervision, nonlin_first: bool = False):

        super().__init__()
        self.deep_supervision = deep_supervision
        self.encoder = encoder
        self.num_classes = num_classes
        n_stages_encoder = len(encoder.output_channels)
        if isinstance(n_conv_per_stage, int):
            n_conv_per_stage = [n_conv_per_stage] * (n_stages_encoder - 1)
        assert len(n_conv_per_stage) == n_stages_encoder - 1, "n_conv_per_stage must have as many entries as we have " \
                                                          "resolution stages - 1 (n_stages in encoder - 1), " \
                                                          "here: %d" % n_stages_encoder
        stages = []

        upsample_layers = []

        seg_layers = []
        for s in range(1, n_stages_encoder):
            input_features_below = encoder.output_channels[-s]
            input_features_skip = encoder.output_channels[-(s + 1)]
            stride_for_upsampling = encoder.strides[-s]

            upsample_layers.append(UpsampleLayer(
                conv_op = encoder.conv_op,
                input_channels = input_features_below,
                output_channels = input_features_skip,
                pool_op_kernel_size = stride_for_upsampling,
                mode='nearest'
            ))

            stages.append(nn.Sequential(
                BasicResBlock(
                    conv_op = encoder.conv_op,
                    norm_op = encoder.norm_op,
                    norm_op_kwargs = encoder.norm_op_kwargs,
                    nonlin = encoder.nonlin,
                    nonlin_kwargs = encoder.nonlin_kwargs,
                    input_channels = 2 * input_features_skip if s < n_stages_encoder - 1 else input_features_skip,
                    output_channels = input_features_skip,
                    kernel_size = encoder.kernel_sizes[-(s + 1)],
                    padding=encoder.conv_pad_sizes[-(s + 1)],
                    stride=1,
                    use_1x1conv=True
                ),
                *[
                    BasicBlockD(
                        conv_op = encoder.conv_op,
                        input_channels = input_features_skip,
                        output_channels = input_features_skip,
                        kernel_size = encoder.kernel_sizes[-(s + 1)],
                        stride = 1,
                        conv_bias = encoder.conv_bias,
                        norm_op = encoder.norm_op,
                        norm_op_kwargs = encoder.norm_op_kwargs,
                        nonlin = encoder.nonlin,
                        nonlin_kwargs = encoder.nonlin_kwargs,
                    ) for _ in range(n_conv_per_stage[s-1] - 1)
                ]
            ))
            seg_layers.append(encoder.conv_op(input_features_skip, num_classes, 1, 1, 0, bias=True))

        self.stages = nn.ModuleList(stages)
        self.upsample_layers = nn.ModuleList(upsample_layers)
        self.seg_layers = nn.ModuleList(seg_layers)

    def forward(self, skips):
        lres_input = skips[-1]
        seg_outputs = []
        for s in range(len(self.stages)):
            x = self.upsample_layers[s](lres_input)
            if s < (len(self.stages) - 1):
                x = torch.cat((x, skips[-(s+2)]), 1)
            x = self.stages[s](x)
            if self.deep_supervision:
                seg_outputs.append(self.seg_layers[s](x))
            elif s == (len(self.stages) - 1):
                seg_outputs.append(self.seg_layers[-1](x))
            lres_input = x

        seg_outputs = seg_outputs[::-1]

        if not self.deep_supervision:
            r = seg_outputs[0]
        else:
            r = seg_outputs
        return r

    def compute_conv_feature_map_size(self, input_size):
        skip_sizes = []
        for s in range(len(self.encoder.strides) - 1):
            skip_sizes.append([i // j for i, j in zip(input_size, self.encoder.strides[s])])
            input_size = skip_sizes[-1]

        assert len(skip_sizes) == len(self.stages)

        output = np.int64(0)
        for s in range(len(self.stages)):
            output += self.stages[s].compute_conv_feature_map_size(skip_sizes[-(s+1)])
            output += np.prod([self.encoder.output_channels[-(s+2)], *skip_sizes[-(s+1)]], dtype=np.int64)
            if self.deep_supervision or (s == (len(self.stages) - 1)):
                output += np.prod([self.num_classes, *skip_sizes[-(s+1)]], dtype=np.int64)
        return output

@MODELS.register_module('UMambaEnc_2d')
class UMambaEnc(nn.Module):
    def __init__(self,
                 input_size: Tuple[int, ...],
                 input_channels: int,
                 n_stages: int,
                 features_per_stage: Union[int, List[int], Tuple[int, ...]],
                 conv_op: Type[_ConvNd],
                 kernel_sizes: Union[int, List[int], Tuple[int, ...]],
                 strides: Union[int, List[int], Tuple[int, ...]],
                 n_conv_per_stage: Union[int, List[int], Tuple[int, ...]],
                 num_classes: int,
                 n_conv_per_stage_decoder: Union[int, Tuple[int, ...], List[int]],
                 conv_bias: bool = False,
                 norm_op: Union[None, Type[nn.Module]] = None,
                 norm_op_kwargs: dict = None,
                 dropout_op: Union[None, Type[_DropoutNd]] = None,
                 dropout_op_kwargs: dict = None,
                 nonlin: Union[None, Type[torch.nn.Module]] = None,
                 nonlin_kwargs: dict = None,
                 deep_supervision: bool = False,
                 stem_channels: int = None
                 ):
        super().__init__()
        n_blocks_per_stage = n_conv_per_stage
        if isinstance(n_blocks_per_stage, int):
            n_blocks_per_stage = [n_blocks_per_stage] * n_stages
        if isinstance(n_conv_per_stage_decoder, int):
            n_conv_per_stage_decoder = [n_conv_per_stage_decoder] * (n_stages - 1)
    
        for s in range(math.ceil(n_stages / 2), n_stages):
            n_blocks_per_stage[s] = 1    

        for s in range(math.ceil((n_stages - 1) / 2 + 0.5), n_stages - 1):
            n_conv_per_stage_decoder[s] = 1


        assert len(n_blocks_per_stage) == n_stages, "n_blocks_per_stage must have as many entries as we have " \
                                                  f"resolution stages. here: {n_stages}. " \
                                                  f"n_blocks_per_stage: {n_blocks_per_stage}"
        assert len(n_conv_per_stage_decoder) == (n_stages - 1), "n_conv_per_stage_decoder must have one less entries " \
                                                                f"as we have resolution stages. here: {n_stages} " \
                                                                f"stages, so it should have {n_stages - 1} entries. " \
                                                                f"n_conv_per_stage_decoder: {n_conv_per_stage_decoder}"
        self.encoder = ResidualMambaEncoder(
            input_size,
            input_channels,
            n_stages,
            features_per_stage,
            conv_op,
            kernel_sizes,
            strides,
            n_blocks_per_stage,
            conv_bias,
            norm_op,
            norm_op_kwargs,
            nonlin,
            nonlin_kwargs,
            return_skips=True,
            stem_channels=stem_channels
        )

        self.decoder = UNetResDecoder(self.encoder, num_classes, n_conv_per_stage_decoder, deep_supervision)

    def forward(self, x):
        skips = self.encoder(x)
        return self.decoder(skips)

    def compute_conv_feature_map_size(self, input_size):
        assert len(input_size) == convert_conv_op_to_dim(self.encoder.conv_op), "just give the image size without color/feature channels or " \
                                                                                "batch channel. Do not give input_size=(b, c, x, y(, z)). " \
                                                                                "Give input_size=(x, y(, z))!"
        return self.encoder.compute_conv_feature_map_size(input_size) + self.decoder.compute_conv_feature_map_size(input_size)


def get_umamba_enc_2d_from_plans(
        plans_manager: PlansManager,
        dataset_json: dict,
        configuration_manager: ConfigurationManager,
        num_input_channels: int,
        deep_supervision: bool = True
    ):
    num_stages = len(configuration_manager.conv_kernel_sizes)

    dim = len(configuration_manager.conv_kernel_sizes[0])
    conv_op = convert_dim_to_conv_op(dim)

    label_manager = plans_manager.get_label_manager(dataset_json)

    segmentation_network_class_name = 'UMambaEnc'
    network_class = UMambaEnc
    kwargs = {
        'UMambaEnc': {
            'input_size': configuration_manager.patch_size,
            'conv_bias': True,
            'norm_op': get_matching_instancenorm(conv_op),
            'norm_op_kwargs': {'eps': 1e-5, 'affine': True},
            'dropout_op': None, 'dropout_op_kwargs': None,
            'nonlin': nn.LeakyReLU, 'nonlin_kwargs': {'inplace': True},
        }
    }

    conv_or_blocks_per_stage = {
        'n_conv_per_stage': configuration_manager.n_conv_per_stage_encoder,
        'n_conv_per_stage_decoder': configuration_manager.n_conv_per_stage_decoder
    }

    model = network_class(
        input_channels=num_input_channels,
        n_stages=num_stages,
        features_per_stage=[min(configuration_manager.UNet_base_num_features * 2 ** i,
                                configuration_manager.unet_max_num_features) for i in range(num_stages)],
        conv_op=conv_op,
        kernel_sizes=configuration_manager.conv_kernel_sizes,
        strides=configuration_manager.pool_op_kernel_sizes,
        num_classes=label_manager.num_segmentation_heads,
        deep_supervision=deep_supervision,
        **conv_or_blocks_per_stage,
        **kwargs[segmentation_network_class_name]
    )
    model.apply(InitWeights_He(1e-2))

    return model


import logging

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from pathlib import Path
from fairscale.nn import checkpoint_wrapper
from mmengine.runner import load_checkpoint
# from mmcv_custom import load_checkpoint
from mmseg.registry import MODELS as BACKBONES
# from mmseg.utils import get_root_logger
from mmengine.logging import MMLogger
from timm.models.layers import LayerNorm2d

from collections import OrderedDict
from typing import Optional, Union, Tuple

from einops.layers.torch import Rearrange
from einops import rearrange
from timm.models.layers import DropPath, trunc_normal_


# https://github.com/chenller/mmseg-extension
class TopkRouting(nn.Module):
    """
    differentiable topk routing with scaling
    Args:
        qk_dim: int, feature dimension of query and key
        topk: int, the 'topk'
        qk_scale: int or None, temperature (multiply) of softmax activation
        with_param: bool, wether inorporate learnable params in routing unit
        diff_routing: bool, wether make routing differentiable
        soft_routing: bool, wether make output value multiplied by routing weights
    """

    def __init__(self, qk_dim, topk=4, qk_scale=None, param_routing=False, diff_routing=False):
        super().__init__()
        self.topk = topk
        self.qk_dim = qk_dim
        self.scale = qk_scale or qk_dim ** -0.5
        self.diff_routing = diff_routing
        # TODO: norm layer before/after linear?
        self.emb = nn.Linear(qk_dim, qk_dim) if param_routing else nn.Identity()
        # routing activation
        self.routing_act = nn.Softmax(dim=-1)

    def forward(self, query: Tensor, key: Tensor) -> Tuple[Tensor]:
        """
        Args:
            q, k: (n, p^2, c) tensor
        Return:
            r_weight, topk_index: (n, p^2, topk) tensor
        """
        if not self.diff_routing:
            query, key = query.detach(), key.detach()
        query_hat, key_hat = self.emb(query), self.emb(key)  # per-window pooling -> (n, p^2, c)
        attn_logit = (query_hat * self.scale) @ key_hat.transpose(-2, -1)  # (n, p^2, p^2)
        topk_attn_logit, topk_index = torch.topk(attn_logit, k=self.topk, dim=-1)  # (n, p^2, k), (n, p^2, k)
        r_weight = self.routing_act(topk_attn_logit)  # (n, p^2, k)

        return r_weight, topk_index


class KVGather(nn.Module):
    def __init__(self, mul_weight='none'):
        super().__init__()
        assert mul_weight in ['none', 'soft', 'hard']
        self.mul_weight = mul_weight

    def forward(self, r_idx: Tensor, r_weight: Tensor, kv: Tensor):
        """
        r_idx: (n, p^2, topk) tensor
        r_weight: (n, p^2, topk) tensor
        kv: (n, p^2, w^2, c_kq+c_v)

        Return:
            (n, p^2, topk, w^2, c_kq+c_v) tensor
        """
        # select kv according to routing index
        n, p2, w2, c_kv = kv.size()
        topk = r_idx.size(-1)
        # print(r_idx.size(), r_weight.size())
        # FIXME: gather consumes much memory (topk times redundancy), write cuda kernel?
        topk_kv = torch.gather(kv.view(n, 1, p2, w2, c_kv).expand(-1, p2, -1, -1, -1),
                               # (n, p^2, p^2, w^2, c_kv) without mem cpy
                               dim=2,
                               index=r_idx.view(n, p2, topk, 1, 1).expand(-1, -1, -1, w2, c_kv)
                               # (n, p^2, k, w^2, c_kv)
                               )

        if self.mul_weight == 'soft':
            topk_kv = r_weight.view(n, p2, topk, 1, 1) * topk_kv  # (n, p^2, k, w^2, c_kv)
        elif self.mul_weight == 'hard':
            raise NotImplementedError('differentiable hard routing TBA')
        # else: #'none'
        #     topk_kv = topk_kv # do nothing

        return topk_kv


class QKVLinear(nn.Module):
    def __init__(self, dim, qk_dim, bias=True):
        super().__init__()
        self.dim = dim
        self.qk_dim = qk_dim
        self.qkv = nn.Linear(dim, qk_dim + qk_dim + dim, bias=bias)

    def forward(self, x):
        q, kv = self.qkv(x).split([self.qk_dim, self.qk_dim + self.dim], dim=-1)
        return q, kv
        # q, k, v = self.qkv(x).split([self.qk_dim, self.qk_dim, self.dim], dim=-1)
        # return q, k, v


class BiLevelRoutingAttention(nn.Module):
    """
    n_win: number of windows in one side (so the actual number of windows is n_win*n_win)
    kv_per_win: for kv_downsample_mode='ada_xxxpool' only, number of key/values per window. Similar to n_win, the actual number is kv_per_win*kv_per_win.
    topk: topk for window filtering
    param_attention: 'qkvo'-linear for q,k,v and o, 'none': param free attention
    param_routing: extra linear for routing
    diff_routing: wether to set routing differentiable
    soft_routing: wether to multiply soft routing weights
    """

    def __init__(self, dim, num_heads=8, n_win=7, qk_dim=None, qk_scale=None,
                 kv_per_win=4, kv_downsample_ratio=4, kv_downsample_kernel=None, kv_downsample_mode='identity',
                 topk=4, param_attention="qkvo", param_routing=False, diff_routing=False, soft_routing=False,
                 side_dwconv=3,
                 auto_pad=False):
        super().__init__()
        # local attention setting
        self.dim = dim
        self.n_win = n_win  # Wh, Ww
        self.num_heads = num_heads
        self.qk_dim = qk_dim or dim
        assert self.qk_dim % num_heads == 0 and self.dim % num_heads == 0, 'qk_dim and dim must be divisible by num_heads!'
        self.scale = qk_scale or self.qk_dim ** -0.5

        ################side_dwconv (i.e. LCE in ShuntedTransformer)###########
        self.lepe = nn.Conv2d(dim, dim, kernel_size=side_dwconv, stride=1, padding=side_dwconv // 2,
                              groups=dim) if side_dwconv > 0 else \
            lambda x: torch.zeros_like(x)

        ################ global routing setting #################
        self.topk = topk
        self.param_routing = param_routing
        self.diff_routing = diff_routing
        self.soft_routing = soft_routing
        # router
        assert not (self.param_routing and not self.diff_routing)  # cannot be with_param=True and diff_routing=False
        self.router = TopkRouting(qk_dim=self.qk_dim,
                                  qk_scale=self.scale,
                                  topk=self.topk,
                                  diff_routing=self.diff_routing,
                                  param_routing=self.param_routing)
        if self.soft_routing:  # soft routing, always diffrentiable (if no detach)
            mul_weight = 'soft'
        elif self.diff_routing:  # hard differentiable routing
            mul_weight = 'hard'
        else:  # hard non-differentiable routing
            mul_weight = 'none'
        self.kv_gather = KVGather(mul_weight=mul_weight)

        # qkv mapping (shared by both global routing and local attention)
        self.param_attention = param_attention
        if self.param_attention == 'qkvo':
            self.qkv = QKVLinear(self.dim, self.qk_dim)
            self.wo = nn.Linear(dim, dim)
        elif self.param_attention == 'qkv':
            self.qkv = QKVLinear(self.dim, self.qk_dim)
            self.wo = nn.Identity()
        else:
            raise ValueError(f'param_attention mode {self.param_attention} is not surpported!')

        self.kv_downsample_mode = kv_downsample_mode
        self.kv_per_win = kv_per_win
        self.kv_downsample_ratio = kv_downsample_ratio
        self.kv_downsample_kenel = kv_downsample_kernel
        if self.kv_downsample_mode == 'ada_avgpool':
            assert self.kv_per_win is not None
            self.kv_down = nn.AdaptiveAvgPool2d(self.kv_per_win)
        elif self.kv_downsample_mode == 'ada_maxpool':
            assert self.kv_per_win is not None
            self.kv_down = nn.AdaptiveMaxPool2d(self.kv_per_win)
        elif self.kv_downsample_mode == 'maxpool':
            assert self.kv_downsample_ratio is not None
            self.kv_down = nn.MaxPool2d(self.kv_downsample_ratio) if self.kv_downsample_ratio > 1 else nn.Identity()
        elif self.kv_downsample_mode == 'avgpool':
            assert self.kv_downsample_ratio is not None
            self.kv_down = nn.AvgPool2d(self.kv_downsample_ratio) if self.kv_downsample_ratio > 1 else nn.Identity()
        elif self.kv_downsample_mode == 'identity':  # no kv downsampling
            self.kv_down = nn.Identity()
        elif self.kv_downsample_mode == 'fracpool':
            # assert self.kv_downsample_ratio is not None
            # assert self.kv_downsample_kenel is not None
            # TODO: fracpool
            # 1. kernel size should be input size dependent
            # 2. there is a random factor, need to avoid independent sampling for k and v
            raise NotImplementedError('fracpool policy is not implemented yet!')
        elif kv_downsample_mode == 'conv':
            # TODO: need to consider the case where k != v so that need two downsample modules
            raise NotImplementedError('conv policy is not implemented yet!')
        else:
            raise ValueError(f'kv_down_sample_mode {self.kv_downsaple_mode} is not surpported!')

        # softmax for local attention
        self.attn_act = nn.Softmax(dim=-1)

        self.auto_pad = auto_pad

    def forward(self, x, ret_attn_mask=False):
        """
        x: NHWC tensor

        Return:
            NHWC tensor
        """
        # NOTE: use padding for semantic segmentation
        ###################################################
        if self.auto_pad:
            N, H_in, W_in, C = x.size()

            pad_l = pad_t = 0
            pad_r = (self.n_win - W_in % self.n_win) % self.n_win
            pad_b = (self.n_win - H_in % self.n_win) % self.n_win
            x = F.pad(x, (0, 0,  # dim=-1
                          pad_l, pad_r,  # dim=-2
                          pad_t, pad_b))  # dim=-3
            _, H, W, _ = x.size()  # padded size
        else:
            N, H, W, C = x.size()
            assert H % self.n_win == 0 and W % self.n_win == 0  #
        ###################################################

        # patchify, (n, p^2, w, w, c), keep 2d window as we need 2d pooling to reduce kv size
        x = rearrange(x, "n (j h) (i w) c -> n (j i) h w c", j=self.n_win, i=self.n_win)

        #################qkv projection###################
        # q: (n, p^2, w, w, c_qk)
        # kv: (n, p^2, w, w, c_qk+c_v)
        # NOTE: separte kv if there were memory leak issue caused by gather
        q, kv = self.qkv(x)

        # pixel-wise qkv
        # q_pix: (n, p^2, w^2, c_qk)
        # kv_pix: (n, p^2, h_kv*w_kv, c_qk+c_v)
        q_pix = rearrange(q, 'n p2 h w c -> n p2 (h w) c')
        kv_pix = self.kv_down(rearrange(kv, 'n p2 h w c -> (n p2) c h w'))
        kv_pix = rearrange(kv_pix, '(n j i) c h w -> n (j i) (h w) c', j=self.n_win, i=self.n_win)

        q_win, k_win = q.mean([2, 3]), kv[..., 0:self.qk_dim].mean(
            [2, 3])  # window-wise qk, (n, p^2, c_qk), (n, p^2, c_qk)

        ##################side_dwconv(lepe)##################
        # NOTE: call contiguous to avoid gradient warning when using ddp
        lepe = self.lepe(rearrange(kv[..., self.qk_dim:], 'n (j i) h w c -> n c (j h) (i w)', j=self.n_win,
                                   i=self.n_win).contiguous())
        lepe = rearrange(lepe, 'n c (j h) (i w) -> n (j h) (i w) c', j=self.n_win, i=self.n_win)

        ############ gather q dependent k/v #################

        r_weight, r_idx = self.router(q_win, k_win)  # both are (n, p^2, topk) tensors

        kv_pix_sel = self.kv_gather(r_idx=r_idx, r_weight=r_weight, kv=kv_pix)  # (n, p^2, topk, h_kv*w_kv, c_qk+c_v)
        k_pix_sel, v_pix_sel = kv_pix_sel.split([self.qk_dim, self.dim], dim=-1)
        # kv_pix_sel: (n, p^2, topk, h_kv*w_kv, c_qk)
        # v_pix_sel: (n, p^2, topk, h_kv*w_kv, c_v)

        ######### do attention as normal ####################
        k_pix_sel = rearrange(k_pix_sel, 'n p2 k w2 (m c) -> (n p2) m c (k w2)',
                              m=self.num_heads)  # flatten to BMLC, (n*p^2, m, topk*h_kv*w_kv, c_kq//m) transpose here?
        v_pix_sel = rearrange(v_pix_sel, 'n p2 k w2 (m c) -> (n p2) m (k w2) c',
                              m=self.num_heads)  # flatten to BMLC, (n*p^2, m, topk*h_kv*w_kv, c_v//m)
        q_pix = rearrange(q_pix, 'n p2 w2 (m c) -> (n p2) m w2 c',
                          m=self.num_heads)  # to BMLC tensor (n*p^2, m, w^2, c_qk//m)

        # param-free multihead attention
        attn_weight = (
                              q_pix * self.scale) @ k_pix_sel  # (n*p^2, m, w^2, c) @ (n*p^2, m, c, topk*h_kv*w_kv) -> (n*p^2, m, w^2, topk*h_kv*w_kv)
        attn_weight = self.attn_act(attn_weight)
        out = attn_weight @ v_pix_sel  # (n*p^2, m, w^2, topk*h_kv*w_kv) @ (n*p^2, m, topk*h_kv*w_kv, c) -> (n*p^2, m, w^2, c)
        out = rearrange(out, '(n j i) m (h w) c -> n (j h) (i w) (m c)', j=self.n_win, i=self.n_win,
                        h=H // self.n_win, w=W // self.n_win)

        out = out + lepe
        # output linear
        out = self.wo(out)

        # NOTE: use padding for semantic segmentation
        # crop padded region
        if self.auto_pad and (pad_r > 0 or pad_b > 0):
            out = out[:, :H_in, :W_in, :].contiguous()

        if ret_attn_mask:
            return out, r_weight, r_idx, attn_weight
        else:
            return out


class DWConv(nn.Module):
    def __init__(self, dim=768):
        super(DWConv, self).__init__()
        self.dwconv = nn.Conv2d(dim, dim, 3, 1, 1, bias=True, groups=dim)

    def forward(self, x):
        """
        x: NHWC tensor
        """
        x = x.permute(0, 3, 1, 2)  # NCHW
        x = self.dwconv(x)
        x = x.permute(0, 2, 3, 1)  # NHWC

        return x


class Attention(nn.Module):
    """
    vanilla attention
    """

    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0.):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        # NOTE scale factor was wrong in my original version, can set manually to be compat with prev weights
        self.scale = qk_scale or head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        """
        args:
            x: NHWC tensor
        return:
            NHWC tensor
        """
        _, H, W, _ = x.size()
        x = rearrange(x, 'n h w c -> n (h w) c')

        #######################################
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]  # make torchscript happy (cannot use tensor as tuple)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        #######################################

        x = rearrange(x, 'n (h w) c -> n h w c', h=H, w=W)
        return x


class AttentionLePE(nn.Module):
    """
    vanilla attention
    """

    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0., side_dwconv=5):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        # NOTE scale factor was wrong in my original version, can set manually to be compat with prev weights
        self.scale = qk_scale or head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.lepe = nn.Conv2d(dim, dim, kernel_size=side_dwconv, stride=1, padding=side_dwconv // 2,
                              groups=dim) if side_dwconv > 0 else \
            lambda x: torch.zeros_like(x)

    def forward(self, x):
        """
        args:
            x: NHWC tensor
        return:
            NHWC tensor
        """
        _, H, W, _ = x.size()
        x = rearrange(x, 'n h w c -> n (h w) c')

        #######################################
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]  # make torchscript happy (cannot use tensor as tuple)

        lepe = self.lepe(rearrange(x, 'n (h w) c -> n c h w', h=H, w=W))
        lepe = rearrange(lepe, 'n c h w -> n (h w) c')

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = x + lepe

        x = self.proj(x)
        x = self.proj_drop(x)
        #######################################

        x = rearrange(x, 'n (h w) c -> n h w c', h=H, w=W)
        return x


def get_pe_layer(emb_dim, pe_dim=None, name='none'):
    if name == 'none':
        return nn.Identity()
    # if name == 'sum':
    #     return Summer(PositionalEncodingPermute2D(emb_dim))
    # elif name == 'npe.sin':
    #     return NeuralPE(emb_dim=emb_dim, pe_dim=pe_dim, mode='sin')
    # elif name == 'npe.coord':
    #     return NeuralPE(emb_dim=emb_dim, pe_dim=pe_dim, mode='coord')
    # elif name == 'hpe.conv':
    #     return HybridPE(emb_dim=emb_dim, pe_dim=pe_dim, mode='conv', res_shortcut=True)
    # elif name == 'hpe.dsconv':
    #     return HybridPE(emb_dim=emb_dim, pe_dim=pe_dim, mode='dsconv', res_shortcut=True)
    # elif name == 'hpe.pointconv':
    #     return HybridPE(emb_dim=emb_dim, pe_dim=pe_dim, mode='pointconv', res_shortcut=True)
    else:
        raise ValueError(f'PE name {name} is not surpported!')


class Block(nn.Module):
    def __init__(self, dim, drop_path=0., layer_scale_init_value=-1,
                 num_heads=8, n_win=7, qk_dim=None, qk_scale=None,
                 kv_per_win=4, kv_downsample_ratio=4, kv_downsample_kernel=None, kv_downsample_mode='ada_avgpool',
                 topk=4, param_attention="qkvo", param_routing=False, diff_routing=False, soft_routing=False,
                 mlp_ratio=4, mlp_dwconv=False,
                 side_dwconv=5, before_attn_dwconv=3, pre_norm=True, auto_pad=False):
        super().__init__()
        qk_dim = qk_dim or dim

        # modules
        if before_attn_dwconv > 0:
            self.pos_embed = nn.Conv2d(dim, dim, kernel_size=before_attn_dwconv, padding=1, groups=dim)
        else:
            self.pos_embed = lambda x: 0
        self.norm1 = nn.LayerNorm(dim, eps=1e-6)  # important to avoid attention collapsing
        if topk > 0:
            self.attn = BiLevelRoutingAttention(dim=dim, num_heads=num_heads, n_win=n_win, qk_dim=qk_dim,
                                                qk_scale=qk_scale, kv_per_win=kv_per_win,
                                                kv_downsample_ratio=kv_downsample_ratio,
                                                kv_downsample_kernel=kv_downsample_kernel,
                                                kv_downsample_mode=kv_downsample_mode,
                                                topk=topk, param_attention=param_attention, param_routing=param_routing,
                                                diff_routing=diff_routing, soft_routing=soft_routing,
                                                side_dwconv=side_dwconv,
                                                auto_pad=auto_pad)
        elif topk == -1:
            self.attn = Attention(dim=dim)
        elif topk == -2:
            self.attn = AttentionLePE(dim=dim, side_dwconv=side_dwconv)
        elif topk == 0:
            self.attn = nn.Sequential(Rearrange('n h w c -> n c h w'),  # compatiability
                                      nn.Conv2d(dim, dim, 1),  # pseudo qkv linear
                                      nn.Conv2d(dim, dim, 5, padding=2, groups=dim),  # pseudo attention
                                      nn.Conv2d(dim, dim, 1),  # pseudo out linear
                                      Rearrange('n c h w -> n h w c')
                                      )
        self.norm2 = nn.LayerNorm(dim, eps=1e-6)
        self.mlp = nn.Sequential(nn.Linear(dim, int(mlp_ratio * dim)),
                                 DWConv(int(mlp_ratio * dim)) if mlp_dwconv else nn.Identity(),
                                 nn.GELU(),
                                 nn.Linear(int(mlp_ratio * dim), dim)
                                 )
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        # tricks: layer scale & pre_norm/post_norm
        if layer_scale_init_value > 0:
            self.use_layer_scale = True
            self.gamma1 = nn.Parameter(layer_scale_init_value * torch.ones((dim)), requires_grad=True)
            self.gamma2 = nn.Parameter(layer_scale_init_value * torch.ones((dim)), requires_grad=True)
        else:
            self.use_layer_scale = False
        self.pre_norm = pre_norm

    def forward(self, x):
        """
        x: NCHW tensor
        """
        # conv pos embedding
        x = x + self.pos_embed(x)
        # permute to NHWC tensor for attention & mlp
        x = x.permute(0, 2, 3, 1)  # (N, C, H, W) -> (N, H, W, C)

        # attention & mlp
        if self.pre_norm:
            if self.use_layer_scale:
                x = x + self.drop_path(self.gamma1 * self.attn(self.norm1(x)))  # (N, H, W, C)
                x = x + self.drop_path(self.gamma2 * self.mlp(self.norm2(x)))  # (N, H, W, C)
            else:
                x = x + self.drop_path(self.attn(self.norm1(x)))  # (N, H, W, C)
                x = x + self.drop_path(self.mlp(self.norm2(x)))  # (N, H, W, C)
        else:  # https://kexue.fm/archives/9009
            if self.use_layer_scale:
                x = self.norm1(x + self.drop_path(self.gamma1 * self.attn(x)))  # (N, H, W, C)
                x = self.norm2(x + self.drop_path(self.gamma2 * self.mlp(x)))  # (N, H, W, C)
            else:
                x = self.norm1(x + self.drop_path(self.attn(x)))  # (N, H, W, C)
                x = self.norm2(x + self.drop_path(self.mlp(x)))  # (N, H, W, C)

        # permute back
        x = x.permute(0, 3, 1, 2)  # (N, H, W, C) -> (N, C, H, W)
        return x


class BiFormer(nn.Module):
    def __init__(self, depth=[3, 4, 8, 3], in_channels=3, num_classes=1000, embed_dim=[64, 128, 320, 512],
                 head_dim=64, qk_scale=None, representation_size=None,
                 drop_path_rate=0., drop_rate=0.,
                 use_checkpoint_stages=[],
                 ########
                 n_win=7,
                 kv_downsample_mode='ada_avgpool',
                 kv_per_wins=[2, 2, -1, -1],
                 topks=[8, 8, -1, -1],
                 side_dwconv=5,
                 layer_scale_init_value=-1,
                 qk_dims=[None, None, None, None],
                 param_routing=False, diff_routing=False, soft_routing=False,
                 pre_norm=True,
                 pe=None,
                 pe_stages=[0],
                 before_attn_dwconv=3,
                 auto_pad=False,
                 # -----------------------
                 kv_downsample_kernels=[4, 2, 1, 1],
                 kv_downsample_ratios=[4, 2, 1, 1],  # -> kv_per_win = [2, 2, 2, 1]
                 mlp_ratios=[4, 4, 4, 4],
                 param_attention='qkvo',
                 mlp_dwconv=False):
        """
        Args:
            depth (list): depth of each stage
            img_size (int, tuple): input image size
            in_chans (int): number of input channels
            num_classes (int): number of classes for classification head
            embed_dim (list): embedding dimension of each stage
            head_dim (int): head dimension
            mlp_ratio (int): ratio of mlp hidden dim to embedding dim
            qkv_bias (bool): enable bias for qkv if True
            qk_scale (float): override default qk scale of head_dim ** -0.5 if set
            representation_size (Optional[int]): enable and set representation layer (pre-logits) to this value if set
            drop_rate (float): dropout rate
            attn_drop_rate (float): attention dropout rate
            drop_path_rate (float): stochastic depth rate
            norm_layer (nn.Module): normalization layer
            conv_stem (bool): whether use overlapped patch stem
        """
        super().__init__()
        self.num_classes = num_classes
        self.num_features = self.embed_dim = embed_dim  # num_features for consistency with other models

        ############ downsample layers (patch embeddings) ######################
        self.downsample_layers = nn.ModuleList()
        # NOTE: uniformer uses two 3*3 conv, while in many other transformers this is one 7*7 conv
        stem = nn.Sequential(
            nn.Conv2d(in_channels, embed_dim[0] // 2, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
            nn.BatchNorm2d(embed_dim[0] // 2),
            nn.GELU(),
            nn.Conv2d(embed_dim[0] // 2, embed_dim[0], kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
            nn.BatchNorm2d(embed_dim[0]),
        )
        if (pe is not None) and 0 in pe_stages:
            stem.append(get_pe_layer(emb_dim=embed_dim[0], name=pe))
        if use_checkpoint_stages:
            stem = checkpoint_wrapper(stem)
        self.downsample_layers.append(stem)

        for i in range(3):
            downsample_layer = nn.Sequential(
                nn.Conv2d(embed_dim[i], embed_dim[i + 1], kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
                nn.BatchNorm2d(embed_dim[i + 1])
            )
            if (pe is not None) and i + 1 in pe_stages:
                downsample_layer.append(get_pe_layer(emb_dim=embed_dim[i + 1], name=pe))
            if use_checkpoint_stages:
                downsample_layer = checkpoint_wrapper(downsample_layer)
            self.downsample_layers.append(downsample_layer)
        ##########################################################################

        self.stages = nn.ModuleList()  # 4 feature resolution stages, each consisting of multiple residual blocks
        nheads = [dim // head_dim for dim in qk_dims]
        dp_rates = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depth))]
        cur = 0
        for i in range(4):
            stage = nn.Sequential(
                *[Block(dim=embed_dim[i], drop_path=dp_rates[cur + j],
                        layer_scale_init_value=layer_scale_init_value,
                        topk=topks[i],
                        num_heads=nheads[i],
                        n_win=n_win,
                        qk_dim=qk_dims[i],
                        qk_scale=qk_scale,
                        kv_per_win=kv_per_wins[i],
                        kv_downsample_ratio=kv_downsample_ratios[i],
                        kv_downsample_kernel=kv_downsample_kernels[i],
                        kv_downsample_mode=kv_downsample_mode,
                        param_attention=param_attention,
                        param_routing=param_routing,
                        diff_routing=diff_routing,
                        soft_routing=soft_routing,
                        mlp_ratio=mlp_ratios[i],
                        mlp_dwconv=mlp_dwconv,
                        side_dwconv=side_dwconv,
                        before_attn_dwconv=before_attn_dwconv,
                        pre_norm=pre_norm,
                        auto_pad=auto_pad) for j in range(depth[i])],
            )
            if i in use_checkpoint_stages:
                stage = checkpoint_wrapper(stage)
            self.stages.append(stage)
            cur += depth[i]

        ##########################################################################
        self.norm = nn.BatchNorm2d(embed_dim[-1])
        # Representation layer
        if representation_size:
            self.num_features = representation_size
            self.pre_logits = nn.Sequential(OrderedDict([
                ('fc', nn.Linear(embed_dim, representation_size)),
                ('act', nn.Tanh())
            ]))
        else:
            self.pre_logits = nn.Identity()

        # Classifier head
        self.head = nn.Linear(embed_dim[-1], num_classes) if num_classes > 0 else nn.Identity()
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {'pos_embed', 'cls_token'}

    def get_classifier(self):
        return self.head

    def reset_classifier(self, num_classes, global_pool=''):
        self.num_classes = num_classes
        self.head = nn.Linear(self.embed_dim, num_classes) if num_classes > 0 else nn.Identity()

    def forward_features(self, x):
        for i in range(4):
            x = self.downsample_layers[i](x)  # res = (56, 28, 14, 7), wins = (64, 16, 4, 1)
            x = self.stages[i](x)
        x = self.norm(x)
        x = self.pre_logits(x)
        return x

    def forward(self, x):
        x = self.forward_features(x)
        x = x.flatten(2).mean(-1)
        x = self.head(x)
        return x


@BACKBONES.register_module(name='BiFormer')
class BiFormerSemanticSegmantation(BiFormer):
    def __init__(self, pretrained=None, **kwargs):
        super().__init__(**kwargs)

        # step 1: remove unused segmentation head & norm
        del self.head  # classification head
        del self.norm  # head norm

        # step 2: add extra norms for dense tasks
        self.extra_norms = nn.ModuleList()
        for i in range(4):
            self.extra_norms.append(LayerNorm2d(self.embed_dim[i]))

        # step 3: initialization & load ckpt
        self.apply(self._init_weights)
        self.init_weights(pretrained=pretrained)

        # step 4: convert sync bn, as the batch size is too small in segmentation
        # TODO: check if this is correct
        nn.SyncBatchNorm.convert_sync_batchnorm(self)

    def init_weights(self, pretrained=None):
        if isinstance(pretrained, str):
            logger = MMLogger.get_current_instance()
            # If it is the pre training weight provided by the author
            if Path(pretrained).name in ['biformer_base_best.pth', 'biformer_small_best.pth', 'biformer_tiny_best.pth']:
                p = torch.load(pretrained)
                info = self.load_state_dict(p['model'], strict=False)
                missing_keys = info[0]
                unexpected_keys = info[1]
                logger.log(msg=f'missing_keys={missing_keys}', level=logging.WARN)
                logger.log(msg=f'unexpected_keys={unexpected_keys}', level=logging.WARN)
            else:
                load_checkpoint(self, pretrained, map_location='cpu', strict=False, logger=logger)
                print(f'Load pretrained model from {pretrained}')

    def forward_features(self, x: torch.Tensor):
        out = []
        for i in range(4):
            x = self.downsample_layers[i](x)
            x = self.stages[i](x)
            # DONE: check the inconsistency -> no effect on performance
            # in the version before submission:
            # x = self.extra_norms[i](x)
            # out.append(x)
            out.append(self.extra_norms[i](x))
        return tuple(out)

    def forward(self, x: torch.Tensor):
        return self.forward_features(x)


"""
这个脚本描述用于训练神经元图像的网络结构
"""

import torch
from torch import nn
from torch.nn import Module
from torch.nn.parallel import DataParallel
from datetime import datetime
import torch.nn.functional as F
from smz.models.waveunet.DWT_IDWT.DWT_IDWT_layer import DWT_3D, IDWT_3D, DWT_3D_tiny
from smz.registry import MODELS

class Cov3x3_BN(Module):
    def __init__(self, in_channels, out_channels, kernel_size = 3, stride = 1,
                 padding = 1, dilation = 1, groups = 1, bias = True, padding_mode = 'zeros', with_BN = True):
        super(Cov3x3_BN, self).__init__()
        self.with_BN = with_BN
        self.cov = nn.Conv3d(in_channels = in_channels, out_channels = out_channels, kernel_size = kernel_size,
                             stride = stride, padding = padding, dilation = dilation, groups = groups, bias = bias)
        self.bn = nn.BatchNorm3d(num_features = out_channels)
        self.relu = nn.ReLU()

    def forward(self, input):
        if self.with_BN:
            return self.relu(self.bn(self.cov(input)))
        else:
            return self.relu(self.cov(input))


class Threshold_HFC(Module):
    def __init__(self, la = 0.25):
        super(Threshold_HFC, self).__init__()
        self.threshold = nn.Hardshrink(lambd = la)
    def forward(self, LLH, LHL, LHH, HLL, HLH, HHL, HHH):
        return self.threshold(LLH), self.threshold(LHL), self.threshold(LHH), \
               self.threshold(HLL), self.threshold(HLH), self.threshold(HHL), self.threshold(HHH)

@MODELS.register_module()
class Neuron_SegNet(Module):
    """
    构建用于神经元图像分割的 3D SegNet
    它的编解码器两端进行下采样和上采样分别使用 max-pooling 和 max-unpooling
    """
    def __init__(self, num_class = 2, with_BN = True, channel_width = 4):
        super(Neuron_SegNet, self).__init__()
        # 32 * 128 * 128
        self.cov3d_11_en = Cov3x3_BN(in_channels = 1, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_1 = nn.MaxPool3d(kernel_size = 2, return_indices = True)
        # 16 * 64 * 64
        self.cov3d_21_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_2 = nn.MaxPool3d(kernel_size = 2, return_indices = True)
        # 8 * 32 * 32
        self.cov3d_31_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_3 = nn.MaxPool3d(kernel_size = 2, return_indices = True)
        #  4 * 16 * 16
        self.cov3d_41_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_en = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_4 = nn.MaxPool3d(kernel_size = 2, return_indices = True)

        # 2 * 8 * 8
        self.cov3d_51 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_52 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)

        self.upsampling_4 = nn.MaxUnpool3d(kernel_size = 2, stride = 2)
        # 4 * 16 * 16
        self.cov3d_41_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_3 = nn.MaxUnpool3d(kernel_size = 2, stride = 2)
        # 8 * 32 * 32
        self.cov3d_31_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_2 = nn.MaxUnpool3d(kernel_size = 2, stride = 2)
        # 16 * 64 * 64
        self.cov3d_21_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_1 = nn.MaxUnpool3d(kernel_size = 2, stride = 2)
        # 16 * 64 * 64
        self.cov3d_11_de = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_de = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)

        self.cov_final = nn.Conv3d(in_channels = 1 * channel_width, out_channels = num_class, kernel_size = 1)

    def forward(self, input):
        output = self.cov3d_12_en(self.cov3d_11_en(input))
        output, indices_1 = self.downsampling_1(output)

        output = self.cov3d_22_en(self.cov3d_21_en(output))
        output, indices_2 = self.downsampling_2(output)

        output = self.cov3d_32_en(self.cov3d_31_en(output))
        output, indices_3 = self.downsampling_3(output)

        output = self.cov3d_42_en(self.cov3d_41_en(output))
        output, indices_4 = self.downsampling_4(output)

        output = self.cov3d_52(self.cov3d_51(output))

        output = self.upsampling_4(input = output, indices = indices_4)
        output = self.cov3d_42_de(self.cov3d_41_de(output))

        output = self.upsampling_3(input = output, indices = indices_3)
        output = self.cov3d_32_de(self.cov3d_31_de(output))

        output = self.upsampling_2(input = output, indices = indices_2)
        output = self.cov3d_22_de(self.cov3d_21_de(output))

        output = self.upsampling_1(input = output, indices = indices_1)
        output = self.cov3d_12_de(self.cov3d_11_de(output))

        output = self.cov_final(output)

        return output

@MODELS.register_module()
class Neuron_UNet_V1(Module):
    """
    构建用于神经元图像分割的 3D UNet
    它的编码器下采样使用 max-pooling、解码器上采样使用 反卷积
    """
    def __init__(self, num_class = 2, with_BN = True, channel_width = 4):
        super(Neuron_UNet_V1, self).__init__()
        # 32 * 128 * 128
        self.cov3d_11_en = Cov3x3_BN(in_channels = 1, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_1 = nn.MaxPool3d(kernel_size = 2)
        # 16 * 64 * 64
        self.cov3d_21_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_2 = nn.MaxPool3d(kernel_size = 2)
        # 8 * 32 * 32
        self.cov3d_31_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_3 = nn.MaxPool3d(kernel_size = 2)
        #  4 * 16 * 16
        self.cov3d_41_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_en = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_4 = nn.MaxPool3d(kernel_size = 2)

        # 2 * 8 * 8
        self.cov3d_51 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_52 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)

        self.upsampling_4 = nn.ConvTranspose3d(in_channels = 8 * channel_width, out_channels = 8 * channel_width,
                                               kernel_size = 1, stride = 2, output_padding = 1, padding = 0)
        # 4 * 16 * 16
        self.cov3d_41_de = Cov3x3_BN(in_channels = 16 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_3 = nn.ConvTranspose3d(in_channels = 4 * channel_width, out_channels = 4 * channel_width,
                                               kernel_size = 1, stride = 2, output_padding = 1, padding = 0)
        # 8 * 32 * 32
        self.cov3d_31_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_2 = nn.ConvTranspose3d(in_channels = 2 * channel_width, out_channels = 2 * channel_width,
                                               kernel_size = 1, stride = 2, output_padding = 1, padding = 0)
        # 16 * 64 * 64
        self.cov3d_21_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_1 = nn.ConvTranspose3d(in_channels = 1 * channel_width, out_channels = 1 * channel_width,
                                               kernel_size = 1, stride = 2, output_padding = 1, padding = 0)
        # 16 * 64 * 64
        self.cov3d_11_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_de = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)

        self.cov_final = nn.Conv3d(in_channels = 1 * channel_width, out_channels = num_class, kernel_size = 1)

    def forward(self, input):
        output_1 = self.cov3d_12_en(self.cov3d_11_en(input))
        output = self.downsampling_1(output_1)

        output_2 = self.cov3d_22_en(self.cov3d_21_en(output))
        output = self.downsampling_2(output_2)

        output_3 = self.cov3d_32_en(self.cov3d_31_en(output))
        output = self.downsampling_3(output_3)

        output_4 = self.cov3d_42_en(self.cov3d_41_en(output))
        output = self.downsampling_4(output_4)

        output = self.cov3d_52(self.cov3d_51(output))

        output = self.upsampling_4(input = output)
        output = self.cov3d_42_de(self.cov3d_41_de(torch.cat((output, output_4), dim = 1)))

        output = self.upsampling_3(input = output)
        output = self.cov3d_32_de(self.cov3d_31_de(torch.cat((output, output_3), dim = 1)))

        output = self.upsampling_2(input = output)
        output = self.cov3d_22_de(self.cov3d_21_de(torch.cat((output, output_2), dim = 1)))

        output = self.upsampling_1(input = output)
        output = self.cov3d_12_de(self.cov3d_11_de(torch.cat((output, output_1), dim = 1)))

        output = self.cov_final(output)

        return output

@MODELS.register_module()
class Neuron_UNet_V2(Module):
    """
    构建用于神经元图像分割的 3D UNet
    它的编码器下采样使用 步长卷积，解码器上采样使用 线性插值
    """
    def __init__(self, num_class = 2, with_BN = True, channel_width = 4):
        super(Neuron_UNet_V2, self).__init__()
        # 32 * 128 * 128
        self.cov3d_11_en = Cov3x3_BN(in_channels = 1, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_1 = nn.Conv3d(in_channels = 1 * channel_width, out_channels = 1 * channel_width,
                                        kernel_size = 1, stride = 2)
        # 16 * 64 * 64
        self.cov3d_21_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_2 = nn.Conv3d(in_channels = 2 * channel_width, out_channels = 2 * channel_width,
                                        kernel_size = 1, stride = 2)
        # 8 * 32 * 32
        self.cov3d_31_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_3 = nn.Conv3d(in_channels = 4 * channel_width, out_channels = 4 * channel_width,
                                        kernel_size = 1, stride = 2)
        #  4 * 16 * 16
        self.cov3d_41_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_en = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_4 = nn.Conv3d(in_channels = 8 * channel_width, out_channels = 8 * channel_width,
                                        kernel_size = 1, stride = 2)

        # 2 * 8 * 8
        self.cov3d_51 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_52 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        # 4 * 16 * 16
        self.cov3d_41_de = Cov3x3_BN(in_channels = 16 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        # 8 * 32 * 32
        self.cov3d_31_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        # 16 * 64 * 64
        self.cov3d_21_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        # 16 * 64 * 64
        self.cov3d_11_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_de = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)

        self.cov_final = nn.Conv3d(in_channels = 1 * channel_width, out_channels = num_class, kernel_size = 1)

    def forward(self, input):
        output_1 = self.cov3d_12_en(self.cov3d_11_en(input))
        output = self.downsampling_1(output_1)

        output_2 = self.cov3d_22_en(self.cov3d_21_en(output))
        output = self.downsampling_2(output_2)

        output_3 = self.cov3d_32_en(self.cov3d_31_en(output))
        output = self.downsampling_3(output_3)

        output_4 = self.cov3d_42_en(self.cov3d_41_en(output))
        output = self.downsampling_4(output_4)

        output = self.cov3d_52(self.cov3d_51(output))

        _, _, D, H, W = output.size()
        output = F.interpolate(output, size = (2 * D, 2 * H, 2 * W), mode = 'trilinear', align_corners = True)
        output = self.cov3d_42_de(self.cov3d_41_de(torch.cat((output, output_4), dim = 1)))

        _, _, D, H, W = output.size()
        output = F.interpolate(output, size = (2 * D, 2 * H, 2 * W), mode = 'trilinear', align_corners = True)
        output = self.cov3d_32_de(self.cov3d_31_de(torch.cat((output, output_3), dim = 1)))

        _, _, D, H, W = output.size()
        output = F.interpolate(output, size = (2 * D, 2 * H, 2 * W), mode = 'trilinear', align_corners = True)
        output = self.cov3d_22_de(self.cov3d_21_de(torch.cat((output, output_2), dim = 1)))

        _, _, D, H, W = output.size()
        output = F.interpolate(output, size = (2 * D, 2 * H, 2 * W), mode = 'trilinear', align_corners = True)
        output = self.cov3d_12_de(self.cov3d_11_de(torch.cat((output, output_1), dim = 1)))

        output = self.cov_final(output)

        return output

@MODELS.register_module()
class Neuron_WaveSNet_V1(Module):
    """
    构建集成小波变换用于神经元分割的 3D UNet
    它的编码器下采样使用 3D-DWT，解码器上采样使用 线性插值
    """
    def __init__(self, num_class = 2, with_BN = True, channel_width = 4, wavename = 'haar'):
        super(Neuron_WaveSNet_V1, self).__init__()
        # 32 * 128 * 128
        self.cov3d_11_en = Cov3x3_BN(in_channels = 1, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_1 = DWT_3D_tiny(wavename = wavename)
        # 16 * 64 * 64
        self.cov3d_21_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_2 = DWT_3D_tiny(wavename = wavename)
        # 8 * 32 * 32
        self.cov3d_31_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_3 = DWT_3D_tiny(wavename = wavename)
        #  4 * 16 * 16
        self.cov3d_41_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_en = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_4 = DWT_3D_tiny(wavename = wavename)

        # 2 * 8 * 8
        self.cov3d_51 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_52 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        # 4 * 16 * 16
        self.cov3d_41_de = Cov3x3_BN(in_channels = 16 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        # 8 * 32 * 32
        self.cov3d_31_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        # 16 * 64 * 64
        self.cov3d_21_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        # 16 * 64 * 64
        self.cov3d_11_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_de = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)

        self.cov_final = nn.Conv3d(in_channels = 1 * channel_width, out_channels = num_class, kernel_size = 1)

    def forward(self, input):
        output_1 = self.cov3d_12_en(self.cov3d_11_en(input))
        output = self.downsampling_1(output_1)

        output_2 = self.cov3d_22_en(self.cov3d_21_en(output))
        output = self.downsampling_2(output_2)

        output_3 = self.cov3d_32_en(self.cov3d_31_en(output))
        output = self.downsampling_3(output_3)

        output_4 = self.cov3d_42_en(self.cov3d_41_en(output))
        output = self.downsampling_4(output_4)

        output = self.cov3d_52(self.cov3d_51(output))

        _, _, D, H, W = output.size()
        output = F.interpolate(output, size = (2 * D, 2 * H, 2 * W), mode = 'trilinear', align_corners = True)
        output = self.cov3d_42_de(self.cov3d_41_de(torch.cat((output, output_4), dim = 1)))

        _, _, D, H, W = output.size()
        output = F.interpolate(output, size = (2 * D, 2 * H, 2 * W), mode = 'trilinear', align_corners = True)
        output = self.cov3d_32_de(self.cov3d_31_de(torch.cat((output, output_3), dim = 1)))

        _, _, D, H, W = output.size()
        output = F.interpolate(output, size = (2 * D, 2 * H, 2 * W), mode = 'trilinear', align_corners = True)
        output = self.cov3d_22_de(self.cov3d_21_de(torch.cat((output, output_2), dim = 1)))

        _, _, D, H, W = output.size()
        output = F.interpolate(output, size = (2 * D, 2 * H, 2 * W), mode = 'trilinear', align_corners = True)
        output = self.cov3d_12_de(self.cov3d_11_de(torch.cat((output, output_1), dim = 1)))

        output = self.cov_final(output)

        return output

@MODELS.register_module()
class Neuron_WaveSNet_V2(Module):
    """
    构建用于神经元图像分割的 3D UNet
    它的编码器下采样使用 3D DWT、解码器上采样使用 反卷积
    """
    def __init__(self, num_class = 2, with_BN = True, channel_width = 4, wavename = 'haar'):
        super(Neuron_WaveSNet_V2, self).__init__()
        # 32 * 128 * 128
        self.cov3d_11_en = Cov3x3_BN(in_channels = 1, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_1 = DWT_3D_tiny(wavename = wavename)
        # 16 * 64 * 64
        self.cov3d_21_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_2 = DWT_3D_tiny(wavename = wavename)
        # 8 * 32 * 32
        self.cov3d_31_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_3 = DWT_3D_tiny(wavename = wavename)
        #  4 * 16 * 16
        self.cov3d_41_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_en = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_4 = DWT_3D_tiny(wavename = wavename)

        # 2 * 8 * 8
        self.cov3d_51 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_52 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)

        self.upsampling_4 = nn.ConvTranspose3d(in_channels = 8 * channel_width, out_channels = 8 * channel_width,
                                               kernel_size = 1, stride = 2, output_padding = 1, padding = 0)
        # 4 * 16 * 16
        self.cov3d_41_de = Cov3x3_BN(in_channels = 16 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_3 = nn.ConvTranspose3d(in_channels = 4 * channel_width, out_channels = 4 * channel_width,
                                               kernel_size = 1, stride = 2, output_padding = 1, padding = 0)
        # 8 * 32 * 32
        self.cov3d_31_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_2 = nn.ConvTranspose3d(in_channels = 2 * channel_width, out_channels = 2 * channel_width,
                                               kernel_size = 1, stride = 2, output_padding = 1, padding = 0)
        # 16 * 64 * 64
        self.cov3d_21_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_1 = nn.ConvTranspose3d(in_channels = 1 * channel_width, out_channels = 1 * channel_width,
                                               kernel_size = 1, stride = 2, output_padding = 1, padding = 0)
        # 16 * 64 * 64
        self.cov3d_11_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_de = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)

        self.cov_final = nn.Conv3d(in_channels = 1 * channel_width, out_channels = num_class, kernel_size = 1)

    def forward(self, input):
        output_1 = self.cov3d_12_en(self.cov3d_11_en(input))
        output = self.downsampling_1(output_1)

        output_2 = self.cov3d_22_en(self.cov3d_21_en(output))
        output = self.downsampling_2(output_2)

        output_3 = self.cov3d_32_en(self.cov3d_31_en(output))
        output = self.downsampling_3(output_3)

        output_4 = self.cov3d_42_en(self.cov3d_41_en(output))
        output = self.downsampling_4(output_4)

        output = self.cov3d_52(self.cov3d_51(output))

        output = self.upsampling_4(input = output)
        output = self.cov3d_42_de(self.cov3d_41_de(torch.cat((output, output_4), dim = 1)))

        output = self.upsampling_3(input = output)
        output = self.cov3d_32_de(self.cov3d_31_de(torch.cat((output, output_3), dim = 1)))

        output = self.upsampling_2(input = output)
        output = self.cov3d_22_de(self.cov3d_21_de(torch.cat((output, output_2), dim = 1)))

        output = self.upsampling_1(input = output)
        output = self.cov3d_12_de(self.cov3d_11_de(torch.cat((output, output_1), dim = 1)))

        output = self.cov_final(output)

        return output

@MODELS.register_module()
class Neuron_WaveSNet_V3(Module):
    """
    构建用于神经元图像分割的 3D UNet
    它的编码器下采样使用 3D DWT、解码器上采样使用 3D IDWT
    """
    def __init__(self, num_class = 2, with_BN = True, channel_width = 4, wavename = 'haar'):
        super(Neuron_WaveSNet_V3, self).__init__()
        # 32 * 128 * 128
        self.cov3d_11_en = Cov3x3_BN(in_channels = 1, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_1 = DWT_3D(wavename = wavename)
        # 16 * 64 * 64
        self.cov3d_21_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_2 = DWT_3D(wavename = wavename)
        # 8 * 32 * 32
        self.cov3d_31_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_3 = DWT_3D(wavename = wavename)
        #  4 * 16 * 16
        self.cov3d_41_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_en = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_4 = DWT_3D(wavename = wavename)

        # 2 * 8 * 8
        self.cov3d_51 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_52 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)

        self.upsampling_4 = IDWT_3D(wavename = wavename)
        # 4 * 16 * 16
        self.cov3d_41_de = Cov3x3_BN(in_channels = 16 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_3 = IDWT_3D(wavename = wavename)
        # 8 * 32 * 32
        self.cov3d_31_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_2 = IDWT_3D(wavename = wavename)
        # 16 * 64 * 64
        self.cov3d_21_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_1 = IDWT_3D(wavename = wavename)
        # 16 * 64 * 64
        self.cov3d_11_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_de = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)

        self.cov_final = nn.Conv3d(in_channels = 1 * channel_width, out_channels = num_class, kernel_size = 1)

    def forward(self, input):
        output_1 = self.cov3d_12_en(self.cov3d_11_en(input))
        output, LLH_1, LHL_1, LHH_1, HLL_1, HLH_1, HHL_1, HHH_1 = self.downsampling_1(output_1)

        output_2 = self.cov3d_22_en(self.cov3d_21_en(output))
        output, LLH_2, LHL_2, LHH_2, HLL_2, HLH_2, HHL_2, HHH_2 = self.downsampling_2(output_2)

        output_3 = self.cov3d_32_en(self.cov3d_31_en(output))
        output, LLH_3, LHL_3, LHH_3, HLL_3, HLH_3, HHL_3, HHH_3 = self.downsampling_3(output_3)

        output_4 = self.cov3d_42_en(self.cov3d_41_en(output))
        output, LLH_4, LHL_4, LHH_4, HLL_4, HLH_4, HHL_4, HHH_4 = self.downsampling_4(output_4)

        output = self.cov3d_52(self.cov3d_51(output))

        output = self.upsampling_4(output, LLH_4, LHL_4, LHH_4, HLL_4, HLH_4, HHL_4, HHH_4)
        output = self.cov3d_42_de(self.cov3d_41_de(torch.cat((output, output_4), dim = 1)))

        output = self.upsampling_3(output, LLH_3, LHL_3, LHH_3, HLL_3, HLH_3, HHL_3, HHH_3)
        output = self.cov3d_32_de(self.cov3d_31_de(torch.cat((output, output_3), dim = 1)))

        output = self.upsampling_2(output, LLH_2, LHL_2, LHH_2, HLL_2, HLH_2, HHL_2, HHH_2)
        output = self.cov3d_22_de(self.cov3d_21_de(torch.cat((output, output_2), dim = 1)))

        output = self.upsampling_1(output, LLH_1, LHL_1, LHH_1, HLL_1, HLH_1, HHL_1, HHH_1)
        output = self.cov3d_12_de(self.cov3d_11_de(torch.cat((output, output_1), dim = 1)))

        output = self.cov_final(output)

        return output

@MODELS.register_module()
class Neuron_WaveSNet_V4(Module):
    """
    构建用于神经元图像分割的 3D UNet
    它的编码器下采样使用 3D DWT、解码器上采样使用 3D IDWT, 高频分量经过滤波处理
    """
    def __init__(self, num_class = 2, with_BN = True, channel_width = 4, wavename = 'haar', la = 0.25):
        super(Neuron_WaveSNet_V4, self).__init__()
        # 32 * 128 * 128
        self.cov3d_11_en = Cov3x3_BN(in_channels = 1, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_1 = DWT_3D(wavename = wavename)
        self.threshold_1 = Threshold_HFC(la = la)
        # 16 * 64 * 64
        self.cov3d_21_en = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_2 = DWT_3D(wavename = wavename)
        self.threshold_2 = Threshold_HFC(la = la)
        # 8 * 32 * 32
        self.cov3d_31_en = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_3 = DWT_3D(wavename = wavename)
        self.threshold_3 = Threshold_HFC(la = la)
        #  4 * 16 * 16
        self.cov3d_41_en = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_en = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.downsampling_4 = DWT_3D(wavename = wavename)
        self.threshold_4 = Threshold_HFC(la = la)

        # 2 * 8 * 8
        self.cov3d_51 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_52 = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)

        self.upsampling_4 = IDWT_3D(wavename = wavename)
        # 4 * 16 * 16
        self.cov3d_41_de = Cov3x3_BN(in_channels = 16 * channel_width, out_channels = 8 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_42_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_3 = IDWT_3D(wavename = wavename)
        # 8 * 32 * 32
        self.cov3d_31_de = Cov3x3_BN(in_channels = 8 * channel_width, out_channels = 4 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_32_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_2 = IDWT_3D(wavename = wavename)
        # 16 * 64 * 64
        self.cov3d_21_de = Cov3x3_BN(in_channels = 4 * channel_width, out_channels = 2 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_22_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.upsampling_1 = IDWT_3D(wavename = wavename)
        # 16 * 64 * 64
        self.cov3d_11_de = Cov3x3_BN(in_channels = 2 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)
        self.cov3d_12_de = Cov3x3_BN(in_channels = 1 * channel_width, out_channels = 1 * channel_width, kernel_size = 3, padding = 1, with_BN = with_BN)

        self.cov_final = nn.Conv3d(in_channels = 1 * channel_width, out_channels = num_class, kernel_size = 1)

    def forward(self, input):
        output_1 = self.cov3d_12_en(self.cov3d_11_en(input))
        output, LLH_1, LHL_1, LHH_1, HLL_1, HLH_1, HHL_1, HHH_1 = self.downsampling_1(output_1)
        LLH_1, LHL_1, LHH_1, HLL_1, HLH_1, HHL_1, HHH_1 = self.threshold_1(LLH_1, LHL_1, LHH_1, HLL_1, HLH_1, HHL_1, HHH_1)

        output_2 = self.cov3d_22_en(self.cov3d_21_en(output))
        output, LLH_2, LHL_2, LHH_2, HLL_2, HLH_2, HHL_2, HHH_2 = self.downsampling_2(output_2)
        LLH_2, LHL_2, LHH_2, HLL_2, HLH_2, HHL_2, HHH_2 = self.threshold_2(LLH_2, LHL_2, LHH_2, HLL_2, HLH_2, HHL_2, HHH_2)

        output_3 = self.cov3d_32_en(self.cov3d_31_en(output))
        output, LLH_3, LHL_3, LHH_3, HLL_3, HLH_3, HHL_3, HHH_3 = self.downsampling_3(output_3)
        LLH_3, LHL_3, LHH_3, HLL_3, HLH_3, HHL_3, HHH_3 = self.threshold_3(LLH_3, LHL_3, LHH_3, HLL_3, HLH_3, HHL_3, HHH_3)

        output_4 = self.cov3d_42_en(self.cov3d_41_en(output))
        output, LLH_4, LHL_4, LHH_4, HLL_4, HLH_4, HHL_4, HHH_4 = self.downsampling_4(output_4)
        LLH_4, LHL_4, LHH_4, HLL_4, HLH_4, HHL_4, HHH_4 = self.threshold_4(LLH_4, LHL_4, LHH_4, HLL_4, HLH_4, HHL_4, HHH_4)

        output = self.cov3d_52(self.cov3d_51(output))

        output = self.upsampling_4(output, LLH_4, LHL_4, LHH_4, HLL_4, HLH_4, HHL_4, HHH_4)
        output = self.cov3d_42_de(self.cov3d_41_de(torch.cat((output, output_4), dim = 1)))

        output = self.upsampling_3(output, LLH_3, LHL_3, LHH_3, HLL_3, HLH_3, HHL_3, HHH_3)
        output = self.cov3d_32_de(self.cov3d_31_de(torch.cat((output, output_3), dim = 1)))

        output = self.upsampling_2(output, LLH_2, LHL_2, LHH_2, HLL_2, HLH_2, HHL_2, HHH_2)
        output = self.cov3d_22_de(self.cov3d_21_de(torch.cat((output, output_2), dim = 1)))

        output = self.upsampling_1(output, LLH_1, LHL_1, LHH_1, HLL_1, HLH_1, HHL_1, HHH_1)
        output = self.cov3d_12_de(self.cov3d_11_de(torch.cat((output, output_1), dim = 1)))

        output = self.cov_final(output)

        return output


if __name__ == '__main__':
    import numpy as np
    from pprint import pprint
    input = torch.rand(size = (1, 1, 128, 128, 128)).float().cuda()
    model = Neuron_WaveSNet_V4(channel_width = 8).cuda()
    start = datetime.now()
    output = model(input)
    print(output.shape)
    stop = datetime.now()
    print('tooking {} secs'.format(stop - start))


# ------------------------------------------------------------------------
# CoTr
# ------------------------------------------------------------------------

import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from smz.models.cotr import CNNBackbone
from smz.models.cotr.detrans.DeformableTrans import DeformableTransformer
from smz.models.cotr.detrans.position_encoding import build_position_encoding
from smz.registry import MODELS

class Conv3d_wd(nn.Conv3d):

    def __init__(self, in_channels, out_channels, kernel_size, stride=(1,1,1), padding=(0,0,0), dilation=(1,1,1), groups=1, bias=False):
        super(Conv3d_wd, self).__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, x):
        weight = self.weight
        weight_mean = weight.mean(dim=1, keepdim=True).mean(dim=2, keepdim=True).mean(dim=3, keepdim=True).mean(dim=4, keepdim=True)
        weight = weight - weight_mean
        # std = weight.view(weight.size(0), -1).std(dim=1).view(-1, 1, 1, 1, 1) + 1e-5
        std = torch.sqrt(torch.var(weight.view(weight.size(0), -1), dim=1) + 1e-12).view(-1, 1, 1, 1, 1)
        weight = weight / std.expand_as(weight)
        return F.conv3d(x, weight, self.bias, self.stride, self.padding, self.dilation, self.groups)


def conv3x3x3(in_planes, out_planes, kernel_size, stride=(1, 1, 1), padding=(0, 0, 0), dilation=(1, 1, 1), groups=1, bias=False, weight_std=False):
    "3x3x3 convolution with padding"
    if weight_std:
        return Conv3d_wd(in_planes, out_planes, kernel_size=kernel_size, stride=stride, padding=padding, dilation=dilation, groups=groups, bias=bias)
    else:
        return nn.Conv3d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, padding=padding, dilation=dilation, groups=groups, bias=bias)


def Norm_layer(norm_cfg, inplanes):

    if norm_cfg == 'BN':
        out = nn.BatchNorm3d(inplanes)
    elif norm_cfg == 'SyncBN':
        out = nn.SyncBatchNorm(inplanes)
    elif norm_cfg == 'GN':
        out = nn.GroupNorm(16, inplanes)
    elif norm_cfg == 'IN':
        out = nn.InstanceNorm3d(inplanes,affine=True)

    return out


def Activation_layer(activation_cfg, inplace=True):

    if activation_cfg == 'ReLU':
        out = nn.ReLU(inplace=inplace)
    elif activation_cfg == 'LeakyReLU':
        out = nn.LeakyReLU(negative_slope=1e-2, inplace=inplace)

    return out


class Conv3dBlock(nn.Module):
    def __init__(self,in_channels,out_channels,norm_cfg,activation_cfg,kernel_size,stride=(1, 1, 1),padding=(0, 0, 0),dilation=(1, 1, 1),bias=False,weight_std=False):
        super(Conv3dBlock,self).__init__()
        self.conv = conv3x3x3(in_channels, out_channels, kernel_size=kernel_size, stride=stride, padding=padding, dilation=dilation, bias=bias, weight_std=weight_std)
        self.norm = Norm_layer(norm_cfg, out_channels)
        self.nonlin = Activation_layer(activation_cfg, inplace=True)
    def forward(self,x):
        x = self.conv(x)
        x = self.norm(x)
        x = self.nonlin(x)
        return x

class ResBlock(nn.Module):

    def __init__(self, inplanes, planes, norm_cfg, activation_cfg, weight_std=False):
        super(ResBlock, self).__init__()
        self.resconv1 = Conv3dBlock(inplanes, planes, norm_cfg, activation_cfg, kernel_size=3, stride=1, padding=1, bias=False, weight_std=weight_std)
        self.resconv2 = Conv3dBlock(planes, planes, norm_cfg, activation_cfg, kernel_size=3, stride=1, padding=1, bias=False, weight_std=weight_std)

    def forward(self, x):
        residual = x

        out = self.resconv1(x)
        out = self.resconv2(out)
        out = out + residual

        return out

@MODELS.register_module('CoTr')
class U_ResTran3D(nn.Module):
    def __init__(self, norm_cfg='BN', activation_cfg='ReLU', img_size=None, num_classes=None, weight_std=False):
        super(U_ResTran3D, self).__init__()

        self.MODEL_NUM_CLASSES = num_classes

        self.upsamplex2 = nn.Upsample(scale_factor=(1,2,2), mode='trilinear')

        self.transposeconv_stage2 = nn.ConvTranspose3d(384, 384, kernel_size=(2,2,2), stride=(2,2,2), bias=False)
        self.transposeconv_stage1 = nn.ConvTranspose3d(384, 192, kernel_size=(2,2,2), stride=(2,2,2), bias=False)
        self.transposeconv_stage0 = nn.ConvTranspose3d(192, 64, kernel_size=(2,2,2), stride=(2,2,2), bias=False)

        self.stage2_de = ResBlock(384, 384, norm_cfg, activation_cfg, weight_std=weight_std)
        self.stage1_de = ResBlock(192, 192, norm_cfg, activation_cfg, weight_std=weight_std)
        self.stage0_de = ResBlock(64, 64, norm_cfg, activation_cfg, weight_std=weight_std)

        self.ds2_cls_conv = nn.Conv3d(384, self.MODEL_NUM_CLASSES, kernel_size=1)
        self.ds1_cls_conv = nn.Conv3d(192, self.MODEL_NUM_CLASSES, kernel_size=1)
        self.ds0_cls_conv = nn.Conv3d(64, self.MODEL_NUM_CLASSES, kernel_size=1)

        self.cls_conv = nn.Conv3d(64, self.MODEL_NUM_CLASSES, kernel_size=1)

        for m in self.modules():
            if isinstance(m, (nn.Conv3d, Conv3d_wd, nn.ConvTranspose3d)):
                m.weight = nn.init.kaiming_normal_(m.weight, mode='fan_out')
            elif isinstance(m, (nn.BatchNorm3d, nn.SyncBatchNorm, nn.InstanceNorm3d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

        self.backbone = CNNBackbone.Backbone(depth=9, norm_cfg=norm_cfg, activation_cfg=activation_cfg, weight_std=weight_std)
        total = sum([param.nelement() for param in self.backbone.parameters()])
        print('  + Number of Backbone Params: %.2f(e6)' % (total / 1e6))

        self.position_embed = build_position_encoding(mode='v2', hidden_dim=384)
        self.encoder_Detrans = DeformableTransformer(d_model=384, dim_feedforward=1536, dropout=0.1, activation='gelu', num_feature_levels=2, nhead=6, num_encoder_layers=6, enc_n_points=4)
        total = sum([param.nelement() for param in self.encoder_Detrans.parameters()])
        print('  + Number of Transformer Params: %.2f(e6)' % (total / 1e6))

    def posi_mask(self, x):

        x_fea = []
        x_posemb = []
        masks = []
        for lvl, fea in enumerate(x):
            if lvl > 1:
                x_fea.append(fea)
                x_posemb.append(self.position_embed(fea))
                masks.append(torch.zeros((fea.shape[0], fea.shape[2], fea.shape[3], fea.shape[4]), dtype=torch.bool).cuda())

        return x_fea, masks, x_posemb


    def forward(self, inputs):
        # # %%%%%%%%%%%%% CoTr
        x_convs = self.backbone(inputs)
        x_fea, masks, x_posemb = self.posi_mask(x_convs)
        x_trans = self.encoder_Detrans(x_fea, masks, x_posemb)

        # # Single_scale
        # # x = self.transposeconv_stage2(x_trans.transpose(-1, -2).view(x_convs[-1].shape))
        # # skip2 = x_convs[-2]
        # Multi-scale   
        x = self.transposeconv_stage2(x_trans[:, x_fea[0].shape[-3]*x_fea[0].shape[-2]*x_fea[0].shape[-1]::].transpose(-1, -2).view(x_convs[-1].shape)) # x_trans length: 12*24*24+6*12*12=7776
        skip2 = x_trans[:, 0:x_fea[0].shape[-3]*x_fea[0].shape[-2]*x_fea[0].shape[-1]].transpose(-1, -2).view(x_convs[-2].shape)
        
        x = x + skip2
        x = self.stage2_de(x)
        ds2 = self.ds2_cls_conv(x)

        x = self.transposeconv_stage1(x)
        skip1 = x_convs[-3]
        x = x + skip1
        x = self.stage1_de(x)
        ds1 = self.ds1_cls_conv(x)

        x = self.transposeconv_stage0(x)
        skip0 = x_convs[-4]
        x = x + skip0
        x = self.stage0_de(x)
        ds0 = self.ds0_cls_conv(x)


        result = self.upsamplex2(x)
        result = self.cls_conv(result)

        return [result, ds0, ds1, ds2]


# class ResTranUnet(SegmentationNetwork):
#     """
#     ResTran-3D Unet
#     """
#     def __init__(self, norm_cfg='BN', activation_cfg='ReLU', img_size=None, num_classes=None, weight_std=False, deep_supervision=False):
#         super().__init__()
#         self.do_ds = False
#         self.U_ResTran3D = U_ResTran3D(norm_cfg, activation_cfg, img_size, num_classes, weight_std) # U_ResTran3D

#         if weight_std==False:
#             self.conv_op = nn.Conv3d
#         else:
#             self.conv_op = Conv3d_wd
#         if norm_cfg=='BN':
#             self.norm_op = nn.BatchNorm3d
#         if norm_cfg=='SyncBN':
#             self.norm_op = nn.SyncBatchNorm
#         if norm_cfg=='GN':
#             self.norm_op = nn.GroupNorm
#         if norm_cfg=='IN':
#             self.norm_op = nn.InstanceNorm3d
#         self.dropout_op = nn.Dropout3d
#         self.num_classes = num_classes
#         self._deep_supervision = deep_supervision
#         self.do_ds = deep_supervision

#     def forward(self, x):
#         seg_output = self.U_ResTran3D(x)
#         if self._deep_supervision and self.do_ds:
#             return seg_output
#         else:
#             return seg_output[0]


# ------------------------------------------------------------------------
# 3D Deformable Transformer
# ------------------------------------------------------------------------
# Modified from Deformable DETR 
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]

import copy
from typing import Optional, List
import math

import torch
import torch.nn.functional as F
from torch import nn
from torch.nn.init import xavier_uniform_, constant_, normal_
from .ops.modules import MSDeformAttn
from .position_encoding import build_position_encoding

class DeformableTransformer(nn.Module):
    def __init__(self, d_model=256, nhead=8,
                 num_encoder_layers=6, dim_feedforward=1024, dropout=0.1,
                 activation="relu", num_feature_levels=4, enc_n_points=4):
        super().__init__()

        self.d_model = d_model
        self.nhead = nhead

        encoder_layer = DeformableTransformerEncoderLayer(d_model, dim_feedforward,
                                                          dropout, activation,
                                                          num_feature_levels, nhead, enc_n_points)
        self.encoder = DeformableTransformerEncoder(encoder_layer, num_encoder_layers)

        self.level_embed = nn.Parameter(torch.Tensor(num_feature_levels, d_model))

        self._reset_parameters()

    def _reset_parameters(self):
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)
        for m in self.modules():
            if isinstance(m, MSDeformAttn):
                m._reset_parameters()
        normal_(self.level_embed)

    def get_valid_ratio(self, mask):
        _, D, H, W = mask.shape
        valid_D = torch.sum(~mask[:, :, 0, 0], 1)
        valid_H = torch.sum(~mask[:, 0, :, 0], 1)
        valid_W = torch.sum(~mask[:, 0, 0, :], 1)

        valid_ratio_d = valid_D.float() / D
        valid_ratio_h = valid_H.float() / H
        valid_ratio_w = valid_W.float() / W
        valid_ratio = torch.stack([valid_ratio_d, valid_ratio_w, valid_ratio_h], -1)
        return valid_ratio

    def forward(self, srcs, masks, pos_embeds):

        # prepare input for encoder
        src_flatten = []
        mask_flatten = []
        lvl_pos_embed_flatten = []
        spatial_shapes = []
        for lvl, (src, mask, pos_embed) in enumerate(zip(srcs, masks, pos_embeds)):
            bs, c, d, h, w = src.shape
            spatial_shape = (d, h, w)
            spatial_shapes.append(spatial_shape)
            src = src.flatten(2).transpose(1, 2)
            mask = mask.flatten(1)
            pos_embed = pos_embed.flatten(2).transpose(1, 2)
            lvl_pos_embed = pos_embed + self.level_embed[lvl].view(1, 1, -1)
            lvl_pos_embed_flatten.append(lvl_pos_embed)
            src_flatten.append(src)
            mask_flatten.append(mask)
        src_flatten = torch.cat(src_flatten, 1)
        mask_flatten = torch.cat(mask_flatten, 1)
        lvl_pos_embed_flatten = torch.cat(lvl_pos_embed_flatten, 1)
        spatial_shapes = torch.as_tensor(spatial_shapes, dtype=torch.long, device=src_flatten.device)
        level_start_index = torch.cat((spatial_shapes.new_zeros((1, )), spatial_shapes.prod(1).cumsum(0)[:-1]))
        valid_ratios = torch.stack([self.get_valid_ratio(m) for m in masks], 1)

        # encoder
        memory = self.encoder(src_flatten, spatial_shapes, level_start_index, valid_ratios, lvl_pos_embed_flatten, mask_flatten)

        return memory



class DeformableTransformerEncoderLayer(nn.Module):
    def __init__(self,
                 d_model=256, d_ffn=1024,
                 dropout=0.1, activation="relu",
                 n_levels=4, n_heads=8, n_points=4):
        super().__init__()

        # self attention
        self.self_attn = MSDeformAttn(d_model, n_levels, n_heads, n_points)
        self.dropout1 = nn.Dropout(dropout)
        self.norm1 = nn.LayerNorm(d_model)

        # ffn
        self.linear1 = nn.Linear(d_model, d_ffn)
        self.activation = _get_activation_fn(activation)
        self.dropout2 = nn.Dropout(dropout)
        self.linear2 = nn.Linear(d_ffn, d_model)
        self.dropout3 = nn.Dropout(dropout)
        self.norm2 = nn.LayerNorm(d_model)

    @staticmethod
    def with_pos_embed(tensor, pos):
        return tensor if pos is None else tensor + pos

    def forward_ffn(self, src):
        src2 = self.linear2(self.dropout2(self.activation(self.linear1(src))))
        src = src + self.dropout3(src2)
        src = self.norm2(src)
        return src

    def forward(self, src, pos, reference_points, spatial_shapes, level_start_index, padding_mask=None):
        # self attention
        src2 = self.self_attn(self.with_pos_embed(src, pos), reference_points, src, spatial_shapes, level_start_index, padding_mask)
        src = src + self.dropout1(src2)
        src = self.norm1(src)

        # ffn
        src = self.forward_ffn(src)

        return src


class DeformableTransformerEncoder(nn.Module):
    def __init__(self, encoder_layer, num_layers):
        super().__init__()
        self.layers = _get_clones(encoder_layer, num_layers)
        self.num_layers = num_layers

    @staticmethod
    def get_reference_points(spatial_shapes, valid_ratios, device):
        reference_points_list = []
        for lvl, (D_, H_, W_) in enumerate(spatial_shapes):

            ref_d, ref_y, ref_x = torch.meshgrid(torch.linspace(0.5, D_ - 0.5, D_, dtype=torch.float32, device=device),
                                                 torch.linspace(0.5, H_ - 0.5, H_, dtype=torch.float32, device=device),
                                                 torch.linspace(0.5, W_ - 0.5, W_, dtype=torch.float32, device=device))

            ref_d = ref_d.reshape(-1)[None] / (valid_ratios[:, None, lvl, 0] * D_)
            ref_y = ref_y.reshape(-1)[None] / (valid_ratios[:, None, lvl, 2] * H_)
            ref_x = ref_x.reshape(-1)[None] / (valid_ratios[:, None, lvl, 1] * W_)

            ref = torch.stack((ref_d, ref_x, ref_y), -1)   # D W H
            reference_points_list.append(ref)
        reference_points = torch.cat(reference_points_list, 1)
        reference_points = reference_points[:, :, None] * valid_ratios[:, None]
        return reference_points

    def forward(self, src, spatial_shapes, level_start_index, valid_ratios, pos=None, padding_mask=None):
        output = src
        reference_points = self.get_reference_points(spatial_shapes, valid_ratios, device=src.device)
        for _, layer in enumerate(self.layers):
            output = layer(output, pos, reference_points, spatial_shapes, level_start_index, padding_mask)

        return output


def _get_clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for i in range(N)])


def _get_activation_fn(activation):
    """Return an activation function given a string"""
    if activation == "relu":
        return F.relu
    if activation == "gelu":
        return F.gelu
    if activation == "glu":
        return F.glu
    raise RuntimeError(F"activation should be relu/gelu, not {activation}.")





# from nnunet_mednext.network_architecture.mednextv1.MedNextV1 import MedNeXt
from smz.models.mednext.MedNextV1 import MedNeXt
from smz.registry import MODELS

@MODELS.register_module('MedNeXt_small')
def create_mednextv1_small(num_input_channels, num_classes, kernel_size=3, ds=False):

    return MedNeXt(
        in_channels = num_input_channels, 
        n_channels = 32,
        n_classes = num_classes, 
        exp_r=2,                         
        kernel_size=kernel_size,         
        deep_supervision=ds,             
        do_res=True,                     
        do_res_up_down = True,
        block_counts = [2,2,2,2,2,2,2,2,2]
    )

@MODELS.register_module('MedNeXt_base')
def create_mednextv1_base(num_input_channels, num_classes, kernel_size=3, ds=False):

    return MedNeXt(
        in_channels = num_input_channels, 
        n_channels = 32,
        n_classes = num_classes, 
        exp_r=[2,3,4,4,4,4,4,3,2],       
        kernel_size=kernel_size,         
        deep_supervision=ds,             
        do_res=True,                     
        do_res_up_down = True,
        block_counts = [2,2,2,2,2,2,2,2,2]
    )

@MODELS.register_module('MedNeXt_medium')
def create_mednextv1_medium(num_input_channels, num_classes, kernel_size=3, ds=False):

    return MedNeXt(
        in_channels = num_input_channels, 
        n_channels = 32,
        n_classes = num_classes, 
        exp_r=[2,3,4,4,4,4,4,3,2],       
        kernel_size=kernel_size,         
        deep_supervision=ds,             
        do_res=True,                     
        do_res_up_down = True,
        block_counts = [3,4,4,4,4,4,4,4,3],
        checkpoint_style = 'outside_block'
    )

@MODELS.register_module('MedNeXt_large')
def create_mednextv1_large(num_input_channels, num_classes, kernel_size=3, ds=False):

    return MedNeXt(
        in_channels = num_input_channels, 
        n_channels = 32,
        n_classes = num_classes, 
        exp_r=[3,4,8,8,8,8,8,4,3],                          
        kernel_size=kernel_size,                     
        deep_supervision=ds,             
        do_res=True,                     
        do_res_up_down = True,
        block_counts = [3,4,8,8,8,8,8,4,3],
        checkpoint_style = 'outside_block'
    )


def create_mednext_v1(num_input_channels, num_classes, model_id, kernel_size=3,
                      deep_supervision=False):

    model_dict = {
        'S': create_mednextv1_small,
        'B': create_mednextv1_base,
        'M': create_mednextv1_medium,
        'L': create_mednextv1_large,
        }
    
    return model_dict[model_id](
        num_input_channels, num_classes, kernel_size, deep_supervision
        )


if __name__ == "__main__":

    model = create_mednextv1_large(1, 3, 3, False)
    print(model)